MODULE Classes;

IMPORT
  SYSTEM, Trace, W := Windows, F := Files, U := Utilites, G := Graphics;

CONST
  typebyte*    = 0;
  typechar*    = 1;
  typeboolean* = 2;
  typeshortint*= 3;
  typeinteger* = 4;
  typelongint* = 5;
  typereal*    = 6;
  typelongreal*= 7;
  typepointer* = 8;
  typeset*     = 9;
  typestring*  = 10;
  typeobject*  = 11;
  typefont*    = 12;

  noClass*     = 0;
  sysClass*    = 1;
  userClass*   = 2;

  alNone*       = 0;
  alLeft*       = 1;
  alTop*        = 2;
  alRight*      = 3;
  alBottom*     = 4;
  alClient*     = 5;

  mbLeft*       = 1;
  mbRight*      = 2;
  mbMiddle*     = 3;

  ssShift*      = 1;
  ssAlt*        = 2;
  ssCtrl*       = 3;
  ssLeft*       = 4;
  ssRight*      = 5;
  ssMiddle*     = 6;
  ssDouble*     = 7;

VAR
  screen*      : Screen;
  mainform*    : Wincontrol;
  classlist*   : Classlist;
  activeform*  : Longint;
  popuprect*   : W.Rect;

TYPE
  Byte         = W.Byte;
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

  Sortfunc         = PROCEDURE {DELEGATE} (object1, object2: Pointer): Boolean;

  Event*      	   = PROCEDURE {DELEGATE};
  NotifyEvent*     = PROCEDURE {DELEGATE} (sender: Object);
  MouseMoveEvent*  = PROCEDURE {DELEGATE} (sender: Object; shift: Set; x,y: Longint);
  MouseEvent*      = PROCEDURE {DELEGATE} (sender: Object; shift: Set; button,x,y:Longint);
  MouseWheelEvent* = PROCEDURE {DELEGATE} (sender: Object; shift: Set; Delta,x,y: Longint);
  KeyPressEvent*   = PROCEDURE {DELEGATE} (sender: Object; VAR key: Char);
  KeyEvent* 	   = PROCEDURE {DELEGATE} (sender: Object; VAR key: Shortint; shift: Set);
  TimerEvent*      = PROCEDURE {DELEGATE} (Sender: Timer);

  Message* = RECORD
    msg*    : Longint;
    wparam* : Longint;
    lparam* : Longint;
    result* : Longint;
    button* : Shortint;
    dc*     : Longint;
  END;

  Screen* = OBJECT (List)
  VAR
    width*  : Longint;
    height* : Longint;
  END Screen;

  Property* = OBJECT
  VAR
    name*     : Name;
    type*     : Shortint;
    byte*     : Byte;
    char*     : Char;
    boolean*  : Boolean;
    shortint* : Shortint;
    integer*  : Integer;
    longint*  : Longint;
    real*     : Real;
    longreal* : Longreal;
    pointer*  : Pointer;
    set*      : Set;
    string*   : Name;
    object*   : Object;
    font*     : G.Font;

    PROCEDURE & Init*(VAR n: String; t: Shortint);
    BEGIN
      U.Copy(n,name);
      type     := t;
      char     := 0X;
      boolean  := FALSE;
      shortint := 0;
      integer  := 0;
      longint  := 0;
      real     := 0.0;
      longreal := 0.0;
      pointer  := NIL;
      set      := {};
      string   := "";
      object   := NIL;
    END Init;

  END Property;

  Timer* = OBJECT
  VAR
    interval*: Longint;
    delay*   : Longint;
    timerid* : Longint;
    OnTimer* : TimerEvent;

    PROCEDURE & Create*;
    BEGIN
      interval := 1000;
      timerid  := 0;
      delay    := 0;
    END Create;

    PROCEDURE Enable*;
    VAR
      rect: W.Rect;
    BEGIN
      IF timerid = 0 THEN
        delay    := 0;
        timerid := W.SetTimer(0,0,interval,TimerProc);
        IF screen = NIL THEN
          NEW(screen,0);
          W.SystemParametersInfo(W.SPI_GETWORKAREA,0,SYSTEM.ADR(rect),0);
          screen.width  := rect.right - rect.left;
          screen.height := rect.bottom - rect.top;
        END;
        screen.Add(SELF);
      END;
    END Enable;

    PROCEDURE Disable*;
    BEGIN
      IF timerid # 0 THEN
        W.KillTimer (0,timerid);
        timerid := 0;
        screen.Remove(SELF);
      END;
    END Disable;

    PROCEDURE Free*;
    BEGIN
      Disable;
    END Free;

    PROCEDURE Timer*;
    BEGIN
      INC(delay);
      IF OnTimer # NIL THEN OnTimer(SELF) END;
    END Timer;

  END Timer;

  Object* = OBJECT
    PROCEDURE Destroy*;
    BEGIN
      W.Collect;
    END Destroy;
  END Object;

  Persistent* = OBJECT (Object)
  VAR
    name*           : Name;
    tag*            : Longint;
    property*       : Propertylist;

    PROCEDURE & Init*;
    BEGIN
      name := "Persistent";
      tag  := 0;
      NEW(property);
      property.AddString("name",name);
      property.AddLongint("tag",tag);
    END Init;

    PROCEDURE Destroy*;
    VAR
      i: Longint;
    BEGIN
      property.Clear;
      property := NIL;
      Destroy^;
    END Destroy;

  END Persistent;

  Basecontrol* = OBJECT (Persistent)
  VAR
    left*      : Longint;
    top*       : Longint;
    width*     : Longint;
    height*    : Longint;

    PROCEDURE & Init*;
    BEGIN
      Init^;
      name   := "Basecontrol";
      left   := 0;
      top    := 0;
      width  := 0;
      height := 0;
      property.AddLongint("left"  ,left);
      property.AddLongint("top"   ,top);
      property.AddLongint("width" ,width);
      property.AddLongint("height",height)
    END Init;

  END Basecontrol;

  Wincontrol* = OBJECT (Basecontrol)
  VAR
    owner*          : Wincontrol;
    handle*         : Longint;
    color*          : Longint;
    classname*      : Name;
    caption*        : Name;
    visible*        : Boolean;
    enabled*        : Boolean;
    style*          : Set;
    exstyle*        : Set;
    tabstop*        : Boolean;
    align*          : Longint;
    controls*       : List;
    font*           : G.Font;

    wndstyle*       : Set;
    wndparent*      : Longint;
    wndparam*       : Longint;
    idcontrol*      : Longint;
    mousedown*      : Boolean;

    brush           : Longint;
    mouseinside     : Boolean;
    mousehover      : Boolean;

    OnMouseDown*    : MouseEvent;
    OnMouseUp*      : MouseEvent;
    OnMouseMove*    : MouseMoveEvent;
    OnClick*        : NotifyEvent;
    OnDblClick*     : NotifyEvent;
    OnMouseWheel*   : MouseWheelEvent;
    OnMouseLeave*   : NotifyEvent;
    OnMouseEnter*   : NotifyEvent;
    OnMouseHover*   : NotifyEvent;
    OnKeyUp*	    : KeyEvent;
    OnKeyDown*	    : KeyEvent;
    OnKeyPress*	    : KeyPressEvent;

    PROCEDURE & Initctl*(AOwner: Wincontrol);
    VAR
      rect: W.Rect;
    BEGIN
      Init^;
      name      := "Wincontrol";
      classname := "Wincontrol";
      caption   := "o.b.e.r.o.n";
      left      := 0;
      top       := 0;
      width     := 25;
      height    := 25;
      visible   := TRUE;
      enabled   := TRUE;
      tabstop   := FALSE;
      handle    := 0;
      align     := alNone;
      color     := G.clBtnFace;
      property.AddString ("caption",caption);
      property.AddBoolean("tabstop",tabstop);
      property.AddBoolean("visible",visible);
      property.AddBoolean("enabled",enabled);
      property.AddLongint("align"  ,align);
      property.AddLongint("color"  ,color);

      style     := {};
      exstyle   := {};
      wndstyle  := {};
      wndparent := 0;
      idcontrol := 0;
      wndparam  := 0;
      brush     := 0;
      mousedown := FALSE;
      mouseinside := FALSE;
      mousehover  := FALSE;

      NEW(font,"Tahoma");
      property.AddFont("font"  ,font);

      IF AOwner # NIL THEN
        AOwner.Insert(SELF);
        color   := AOwner.color;
      ELSE
        IF screen = NIL THEN
          NEW(screen,0);
          W.SystemParametersInfo(W.SPI_GETWORKAREA,0,SYSTEM.ADR(rect),0);
          screen.width  := rect.right - rect.left;
          screen.height := rect.bottom - rect.top;
        END;
        screen.Add(SELF);
        IF mainform = NIL THEN mainform := SELF END;
      END;
    END Initctl;

    PROCEDURE Destroy*;
    VAR
      i: Longint;
    BEGIN
      IF controls # NIL THEN
        FOR i := 0 TO controls.count - 1 DO
          IF controls.list[i](Wincontrol) # NIL THEN
            controls.list[i](Wincontrol).Destroy;
            classlist.Delete(controls.list[i](Wincontrol).classname);
          END;
        END;
        controls.Clear;
        controls := NIL;
      END;
      classlist.Delete(classname);
      font.Free;
      Destroy^;
    END Destroy;

    PROCEDURE Insert(AControl: Wincontrol);
    BEGIN
      IF controls = NIL THEN NEW(controls,0) END;
      controls.Add(AControl);
      AControl.owner := SELF;
    END Insert;

    PROCEDURE Remove(AControl: Wincontrol);
    BEGIN
      AControl.owner := NIL;
      IF controls # NIL THEN
        controls.Remove(AControl);
        IF controls.count = 0 THEN controls := NIL END;
      END;
    END Remove;

    PROCEDURE UpdateBounds;
    VAR
      R : W.Rect;
    BEGIN
      IF owner = NIL THEN
        W.GetWindowRect(handle,R);
        W.AdjustWindowRectEx(R,{},0,exstyle);
        left := R.left;
        top  := R.top;
      ELSE
        W.GetClientRect(handle,R);
        W.AdjustWindowRectEx(R,style,0,exstyle);
      END;
      width  := R.right-R.left;
      height := R.bottom-R.top;
    END UpdateBounds;

    PROCEDURE CreateParams*;
    BEGIN
      style     := W.WS_CLIPCHILDREN + W.WS_CHILD + W.WS_CLIPSIBLINGS;
      exstyle   := W.WS_EX_CONTROLPARENT;
      IF owner # NIL THEN wndparent := owner.handle END;
    END CreateParams;

    PROCEDURE CreateHandle;
    BEGIN
      handle := W.CreateWindowEx(
        exstyle, classname,
        caption, style,
        left, top, width, height,
        wndparent, idcontrol, W.hInstance,
        wndparam
      );
      IF handle # 0 THEN
        W.SetProp(handle,SYSTEM.ADR("Wincontrol"),SYSTEM.VAL(Longint,SELF));
      END;
    END CreateHandle;

    PROCEDURE RegisterClass;
    CONST
      maxcontrol = 500;
    VAR
      class: W.WndClassEx;
      typ     : Shortint;
      tmpname : Name;
      parent  : Wincontrol;
      res     : Boolean;

      PROCEDURE CreateName;
      VAR
        i: Longint;
      BEGIN
        i := 1;
        tmpname := classname;
        U.Append(U.IntToStr(i),tmpname);
        typ := classlist.Find(tmpname);
        WHILE (typ # noClass) & (i<maxcontrol) DO
          INC(i);
          tmpname := classname;
          U.Append(U.IntToStr(i),tmpname);
          typ := classlist.Find(tmpname);
        END;
        name := tmpname;
      END CreateName;

    BEGIN
      IF ~ enabled THEN style := style + W.WS_DISABLED END;
      IF tabstop   THEN style := style + W.WS_TABSTOP  END;

      typ := classlist.Find(classname);
      IF typ = noClass THEN
        CreateName;
        classname := tmpname;
      ELSIF typ = sysClass THEN
        CreateName;
        classlist.Add(tmpname,TRUE);
        typ := sysClass;
      END;
      IF caption = "o.b.e.r.o.n" THEN caption := name END;

      IF typ = noClass THEN
        IF brush # 0 THEN W.DeleteObject(brush) END;
        brush               := W.CreateSolidBrush(color);
        class.cbSize        := SIZE(W.WndClassEx);
        class.style         := wndstyle;
        class.lpfnWndProc   := WndProc;
        class.cbClsExtra    := 0;
        class.cbWndExtra    := 0;
        class.hInstance     := W.hInstance;
        class.hIcon         := W.LoadIcon(W.hInstance,SYSTEM.VAL(W.LPSTR,1));
        class.hIconSm       := W.LoadIcon(W.hInstance,SYSTEM.VAL(W.LPSTR,1));
        class.hCursor       := W.LoadCursor(0,SYSTEM.VAL(W.LPSTR,W.IDC_ARROW));
        class.hbrBackground := brush;
        class.lpszMenuName  := W.NULL;
        class.lpszClassName := SYSTEM.ADR(classname);
        W.RegisterClassEx(class);
        classlist.Add(classname,FALSE);
      ELSIF typ = sysClass THEN
        IF owner # NIL THEN
          IF owner.controls # NIL THEN
            idcontrol := owner.controls.IndexOf(SELF);
          END;
        END;
      END;
    END RegisterClass;

    PROCEDURE Hide*;
    BEGIN
      visible := FALSE;
    END Hide;

    PROCEDURE Show*;
    BEGIN
      visible := TRUE;
    END Show;

    PROCEDURE BeforeCreate*;
    BEGIN
    END BeforeCreate;

    PROCEDURE AfterCreate*;
    BEGIN
    END AfterCreate;

    PROCEDURE PrecedingWindow(control: Wincontrol): Longint;
    VAR
      i     : Longint;
      Result: Longint;
    BEGIN
      FOR i := controls.IndexOf(control) + 1 TO controls.count - 1 DO
        Result := controls.list[i](Wincontrol).handle;
        IF Result # 0 THEN RETURN Result END;
      END;
      RETURN W.HWND_TOPMOST;
    END PrecedingWindow;

    PROCEDURE CreateWindow(): Boolean;
    BEGIN
      CreateParams;
      RegisterClass;

      IF (wndparent = 0) & (style * W.WS_CHILD # {}) THEN
        W.MessageBox(0,"wndparent = 0 & style = W.WS_CHILD","Error!",W.MB_ICONERROR+W.MB_OK);
        IF owner = NIL THEN
          screen.Remove(SELF);
        ELSE
          owner.Remove(SELF);
        END;
        RETURN FALSE;
      END;

      CreateHandle;

      IF handle # 0 THEN
        UpdateBounds;
        IF owner # NIL THEN
          W.SetWindowPos(
            handle, owner.PrecedingWindow(SELF),
            0, 0, 0, 0,
            W.SWP_NOMOVE + W.SWP_NOSIZE + W.SWP_NOACTIVATE
          );
        END;
      ELSE
        W.MessageBox(0,"handle = 0","Error!",W.MB_ICONERROR+W.MB_OK);
        IF owner = NIL THEN
          screen.Remove(SELF);
        ELSE
          owner.Remove(SELF);
        END;
        RETURN FALSE;
      END;

      RETURN TRUE;
    END CreateWindow;

    PROCEDURE AdjustRect*;
    BEGIN
    END AdjustRect;

    PROCEDURE AlignControls;
    VAR
      i,a         : Longint;
      lw,th,rw,bh : Longint;
      rect        : W.Rect;
      leftrect    : W.Rect;
      toprect     : W.Rect;
      rightrect   : W.Rect;
      bottomrect  : W.Rect;
      clientrect  : W.Rect;

      newleft     : Longint;
      newtop      : Longint;
      newwidth    : Longint;
      newheight   : Longint;

      flag        : Boolean;

      PROCEDURE AdjustRects;
      VAR
        i: Longint;
        a: Longint;
      BEGIN
        leftrect  := rect;
        leftrect.right := leftrect.left;
        toprect   := rect;
        toprect.bottom := toprect.top;
        rightrect := rect;
        rightrect.left := rightrect.right;
        bottomrect:= rect;
        bottomrect.top := bottomrect.bottom;
        clientrect:= rect;

        FOR i := controls.count - 1 TO 0 BY -1 DO
          controls.list[i](Wincontrol).AdjustRect;
          IF controls.list[i](Wincontrol).visible THEN
            a := controls.list[i](Wincontrol).align;
            CASE a OF
            |  alLeft   : INC(leftrect.right,controls.list[i](Wincontrol).width);
            |  alTop    : INC(toprect.bottom,controls.list[i](Wincontrol).height);
            |  alRight  : DEC(rightrect.left,controls.list[i](Wincontrol).width);
            |  alBottom : DEC(bottomrect.top,controls.list[i](Wincontrol).height);
            ELSE
            END;
          END;
        END;

        IF bottomrect.top < toprect.bottom THEN
          bottomrect.bottom := toprect.bottom + bottomrect.bottom - bottomrect.top;
          bottomrect.top := toprect.bottom
        END;
        leftrect.top      := toprect.bottom;
        leftrect.bottom   := bottomrect.top;
        rightrect.top     := toprect.bottom;
        rightrect.bottom  := bottomrect.top;
        IF rightrect.left < leftrect.right THEN
          rightrect.right := leftrect.right + rightrect.right - rightrect.left;
          rightrect.left := leftrect.right;
        END;
        clientrect.left   := leftrect.right;
        clientrect.top    := toprect.bottom;
        clientrect.right  := rightrect.left;
        clientrect.bottom := bottomrect.top;
      END AdjustRects;

    BEGIN
      IF controls # NIL THEN
        W.GetClientRect(handle,rect);

        AdjustRects;
        lw := 0;
        th := 0;
        rw := 0;
        bh := 0;
        flag := FALSE;
        FOR i := 0 TO controls.count - 1 DO
          IF controls.list[i](Wincontrol).visible THEN
            a := controls.list[i](Wincontrol).align;
            CASE a OF
            |  alLeft   :
                          newleft   := leftrect.left + lw;
                          lw        := lw + controls.list[i](Wincontrol).width;
                          newtop    := leftrect.top;
                          newwidth  := controls.list[i](Wincontrol).width;
                          newheight := leftrect.bottom - leftrect.top;
            |  alTop    :
                          newleft   := toprect.left;
                          newtop    := th + toprect.top;
                          th        := th + controls.list[i](Wincontrol).height;
                          newwidth  := toprect.right - toprect.left;
                          newheight := controls.list[i](Wincontrol).height;
            |  alRight  :
                          newleft   := rightrect.right - rw - controls.list[i](Wincontrol).width;
                          rw        := rw + controls.list[i](Wincontrol).width;
                          newtop    := rightrect.top;
                          newwidth  := controls.list[i](Wincontrol).width;
                          newheight := rightrect.bottom - rightrect.top;
            |  alBottom :
                          newleft   := bottomrect.left;
                          newtop    := bottomrect.bottom - bh - controls.list[i](Wincontrol).height;
                          bh        := bh + controls.list[i](Wincontrol).height;
                          newwidth  := bottomrect.right - bottomrect.left;
                          newheight := controls.list[i](Wincontrol).height;
            |  alClient :
                          newleft   := leftrect.right;
                          newtop    := toprect.bottom;
                          newwidth  := rightrect.left - leftrect.right;
                          newheight := bottomrect.top - toprect.bottom;

            ELSE
            END;

            IF controls.list[i](Wincontrol).align # alNone THEN
              controls.list[i](Wincontrol).left   := newleft;
              controls.list[i](Wincontrol).top    := newtop;
              controls.list[i](Wincontrol).width  := newwidth;
              controls.list[i](Wincontrol).height := newheight;
            END;

            controls.list[i](Wincontrol).CheckProperty;
          END;
        END;

      END;

    END AlignControls;

    PROCEDURE CreateControl*;
    VAR
      i,param: Longint;
    BEGIN
      IF handle = 0 THEN
        IF CreateWindow() THEN
          IF controls # NIL THEN
            FOR i := 0 TO controls.count - 1 DO
              controls.list[i](Wincontrol).CreateControl;
            END;
          END;
          IF visible THEN
            param := W.SW_SHOW;
          ELSE
            param := W.SW_HIDE;
          END;
          AlignControls;
          W.ShowWindow(handle,param);
          AfterCreate;
        ELSE
          W.ExitProcess(0);
        END;
      END;
    END CreateControl;

    (*-----------------------------------------------*)

    PROCEDURE MouseDown*(shift: Set; button,x,y: Longint);
    BEGIN
      IF OnMouseDown # NIL THEN OnMouseDown(SELF,shift,button,x,y); END;
    END MouseDown;

    PROCEDURE MouseUp*(shift: Set; button,x,y: Longint);
    BEGIN
      IF OnMouseUp # NIL THEN OnMouseUp(SELF,shift,button,x,y); END;
    END MouseUp;

    PROCEDURE MouseMove*(shift: Set; x,y: Longint);
    BEGIN
      IF OnMouseMove # NIL THEN OnMouseMove(SELF,shift,x,y); END;
    END MouseMove;

    PROCEDURE Click*;
    BEGIN
      IF OnClick # NIL THEN OnClick(SELF); END;
    END Click;

    PROCEDURE DblClick*;
    BEGIN
      IF OnDblClick # NIL THEN OnDblClick(SELF); END;
    END DblClick;

    PROCEDURE MouseWheel*(shift: Set; delta: Longint; x,y: Longint);
    BEGIN
      IF OnMouseWheel # NIL THEN
        OnMouseWheel(SELF,shift,delta,x,y);
      END;
    END MouseWheel;

    PROCEDURE MouseLeave*;
    BEGIN
      IF OnMouseLeave # NIL THEN OnMouseLeave(SELF); END;
    END MouseLeave;

    PROCEDURE MouseEnter*;
    BEGIN
      IF OnMouseEnter # NIL THEN OnMouseEnter(SELF); END;
    END MouseEnter;

    PROCEDURE MouseHover*;
    BEGIN
      IF OnMouseHover # NIL THEN OnMouseHover(SELF); END;
    END MouseHover;

    PROCEDURE KeyUp*(key: Shortint; shift: Set);
    BEGIN
      IF OnKeyUp # NIL THEN OnKeyUp(SELF,key,shift); END;
    END KeyUp;

    PROCEDURE KeyDown*(key: Shortint; shift: Set);
    BEGIN
      IF OnKeyDown # NIL THEN OnKeyDown(SELF,key,shift); END;
    END KeyDown;

    PROCEDURE KeyPress*(key: Char);
    BEGIN
      IF OnKeyPress # NIL THEN OnKeyPress(SELF,key); END;
    END KeyPress;

    PROCEDURE Resize*;
    BEGIN
    END Resize;

    PROCEDURE FontChange*;
    BEGIN
    END FontChange;

    PROCEDURE SetFocus*;
    BEGIN
      W.SetFocus(handle);
    END SetFocus;

    PROCEDURE ReAlign*;
    BEGIN
      AlignControls;
    END ReAlign;

    PROCEDURE Update*;
    BEGIN
    END Update;

   (*-----------------------------------------------*)

    PROCEDURE WMEnable*;
    BEGIN
    END WMEnable;

    PROCEDURE WMDisable*;
    BEGIN
    END WMDisable;

    PROCEDURE WMShow*;
    BEGIN
    END WMShow;

    PROCEDURE WMHide*;
    BEGIN
    END WMHide;

    PROCEDURE WMMouseLeave*;
    BEGIN
      MouseLeave;
    END WMMouseLeave;

    PROCEDURE WMMouseEnter*;
    BEGIN
      MouseEnter;
    END WMMouseEnter;

    PROCEDURE WMMouseHover*;
    BEGIN
      MouseHover;
    END WMMouseHover;

    PROCEDURE WMCommand*(VAR message: Message);
    BEGIN
    END WMCommand;

    PROCEDURE WMMouseMove*(VAR message: Message);
    VAR
      x,y: Longint;
      key: Set;
    BEGIN
      x := U.LoWord(message.lparam);
      y := U.HiWord(message.lparam);
      key := SysKeyToSet(message.wparam);
      MouseMove(key,x,y);
    END WMMouseMove;

    PROCEDURE WMMouseClick*(VAR message: Message);
    BEGIN
      IF message.button = mbLeft THEN
        Click;
      END;
    END WMMouseClick;

    PROCEDURE WMDblClick*(VAR message: Message);
    BEGIN
      DblClick;
    END WMDblClick;

    PROCEDURE WMMouseDown*(VAR message: Message);
    VAR
      x,y: Longint;
      key: Set;
    BEGIN
      x := U.LoWord(message.lparam);
      y := U.HiWord(message.lparam);
      key := SysKeyToSet(message.wparam);
      MouseDown(key,message.button,x,y);
      IF (message.button = mbLeft)  THEN PopupHide(NIL) END;
      IF (message.button = mbRight) THEN PopupHide(SELF) END;
    END WMMouseDown;

    PROCEDURE WMNCMouseDown*(VAR message: Message);
    BEGIN
      PopupHide(NIL);
    END WMNCMouseDown;

    PROCEDURE WMMouseUp*(VAR message: Message);
    VAR
      x,y: Longint;
      key: Set;
    BEGIN
      x := U.LoWord(message.lparam);
      y := U.HiWord(message.lparam);
      key := SysKeyToSet(message.wparam);
      MouseUp(key,message.button,x,y);
    END WMMouseUp;

    PROCEDURE WMMouseWheel*(VAR message: Message);
    VAR
      x,y,delta: Longint;
      key      : Set;
    BEGIN
      x := U.LoWord(message.lparam);
      y := U.HiWord(message.lparam);
      delta := U.HiWord(message.wparam) DIV 120;
      key := SysKeyToSet(message.wparam);
      MouseWheel(key,delta,x,y);
    END WMMouseWheel;

    PROCEDURE WMMove*(VAR message: Message);
    BEGIN
    END WMMove;

    PROCEDURE WMKeyDown*(VAR message: Message);
    BEGIN
      KeyDown(SHORT(SHORT(message.wparam)),SYSTEM.VAL(Set,message.lparam));
    END WMKeyDown;

    PROCEDURE WMKeyUp*(VAR message: Message);
    BEGIN
      KeyUp(SHORT(SHORT(message.wparam)),SYSTEM.VAL(Set,message.lparam));
    END WMKeyUp;

    PROCEDURE WMChar*(VAR message: Message);
    BEGIN
      KeyPress(CHR(SHORT(U.LoWord(message.wparam))));
    END WMChar;

    PROCEDURE WMSize*(VAR message: Message);
    BEGIN
      Resize;
    END WMSize;

    PROCEDURE WMActivate*(VAR message: Message);
    BEGIN
    END WMActivate;

    PROCEDURE WMInActivate*(VAR message: Message);
    VAR
      p      : W.Point;
      rect   : W.Rect;
    BEGIN
      W.GetCursorPos(p);
      W.GetWindowRect(handle,rect);
      IF ~ U.PtInRect(rect,p) & ~ U.PtInRect(popuprect,p) THEN
        PopupHide(NIL);
      END;
    END WMInActivate;

    PROCEDURE WMSetFocus*(VAR message: Message);
    BEGIN
    END WMSetFocus;

    PROCEDURE WMKillFocus*(VAR message: Message);
    BEGIN
    END WMKillFocus;

    PROCEDURE DrawCaption*(dc: W.HDC);
    VAR
      size : W.Size;
      old  : Longint;
      rect : W.Rect;
    BEGIN
      IF owner # NIL THEN
        old := W.SetBkMode(dc,W.TRANSPARENT);
        W.SetTextColor(dc,font.color);
        W.SelectObject(dc, font.handle);
        W.GetTextExtentPoint32(dc,SYSTEM.ADR(caption),U.Length(caption),size);
        rect.left   := (width - size.cx) DIV 2 - 1;
        rect.top    := (height - size.cy) DIV 2;
        rect.right  := rect.left + size.cx + 2;
        rect.bottom := rect.top + size.cy;
        W.FillRect(dc,rect,brush);
        W.ExtTextOut(dc,rect.left+1,rect.top,W.ETO_CLIPPED,NIL,SYSTEM.ADR(caption),U.Length(caption), NIL);
        W.SetBkMode(dc,old);
      END;
    END DrawCaption;

    PROCEDURE WMPaint*(VAR message: Message);
    VAR
      dc    : W.HDC;
    BEGIN
      dc := W.GetDC(handle);
      DrawCaption(dc);
      W.ReleaseDC(handle,dc);
    END WMPaint;

    PROCEDURE WMEraseBkgnd*(VAR message: Message);
    BEGIN
    END WMEraseBkgnd;

    PROCEDURE SysKeyToSet(Param: Longint): Set;
    VAR
      key, sysKey: Set;
    BEGIN
      key := {};
      sysKey := SYSTEM.VAL(Set,Param);
      IF 2 IN sysKey THEN INCL(key,ssShift); END;
      IF 3 IN sysKey THEN INCL(key,ssCtrl); END;
      RETURN key;
    END SysKeyToSet;

    (*----------------------------------------*)

    PROCEDURE CheckProperty*;
    VAR
      i   : Longint;
      n   : Name;
      rect: W.Rect;
    BEGIN

      IF handle # 0 THEN

        IF controls # NIL THEN
          FOR i := 0 TO controls.count - 1 DO
            controls.list[i](Wincontrol).CheckProperty;
          END;
        END;

        FOR i := 0 TO property.count - 1 DO
          n := property.list[i](Property).name;
          CASE n[0] OF
          | "n" : IF (n = "name")   & (name   # property.list[i](Property).string)  THEN INCL(property.modified,1) END;
          | "l" : IF (n = "left")   & (left   # property.list[i](Property).longint) THEN INCL(property.modified,3) END;
          | "t" :
                  IF (n = "top") THEN
                    IF (top    # property.list[i](Property).longint) THEN INCL(property.modified,4) END;
                  ELSIF (n = "tag") THEN
                    IF (tag    # property.list[i](Property).longint) THEN INCL(property.modified,2) END;
                  ELSIF (n = "tabstop") THEN
                    IF (tabstop    # property.list[i](Property).boolean) THEN INCL(property.modified,10) END;
                  END;
          | "w" : IF (n = "width")  & (width  # property.list[i](Property).longint) THEN INCL(property.modified,5) END;
          | "h" : IF (n = "height") & (height # property.list[i](Property).longint) THEN INCL(property.modified,6) END;
          | "v" : IF (n = "visible") & (visible # property.list[i](Property).boolean) THEN INCL(property.modified,7) END;
          | "e" : IF (n = "enabled") & (enabled # property.list[i](Property).boolean) THEN INCL(property.modified,8) END;
          | "c" :
                  IF (n = "caption") THEN
                    IF (caption # property.list[i](Property).string) THEN INCL(property.modified,9) END;
                  ELSIF (n = "color") THEN
                    IF (color    # property.list[i](Property).longint) THEN INCL(property.modified,12) END;
                  END;
          | "a" : IF (n = "align") & (align # property.list[i](Property).longint) THEN INCL(property.modified,11) END;
          | "f" :
                  IF (n = "font") THEN
                    IF (font.name   # property.list[i](Property).font.name)   THEN INCL(property.modified,13) END;
                    IF (font.color  # property.list[i](Property).font.color)  THEN INCL(property.modified,14) END;
                    IF (font.height # property.list[i](Property).font.height) THEN INCL(property.modified,15) END;
                    IF (font.style  # property.list[i](Property).font.style)  THEN INCL(property.modified,16) END;
                  ELSE
                  END;
          ELSE
          END;

        END;

      END;

      IF property.modified # {} THEN

        IF 1 IN property.modified THEN property.AddString ("name",name) END;
        IF 2 IN property.modified THEN property.AddLongint("tag" ,tag)  END;

        IF {3,4,5,6} * property.modified # {} THEN
          IF owner = NIL THEN
            W.GetWindowRect(handle,rect);
          ELSE
            W.GetClientRect(handle,rect);
          END;
          IF (left # rect.left) OR (top # rect.top) OR (width # rect.right - rect.left) OR (height # rect.bottom - rect.top) THEN
            W.MoveWindow(handle,left,top,width,height,1);
          END;
          IF 3 IN property.modified THEN property.AddLongint("left"  ,left) END;
          IF 4 IN property.modified THEN property.AddLongint("top"   ,top) END;
          IF 5 IN property.modified THEN property.AddLongint("width" ,width) END;
          IF 6 IN property.modified THEN property.AddLongint("height",height) END;
        END;

        IF 7 IN property.modified THEN
          IF ~ visible THEN
            W.ShowWindow(handle, W.SW_HIDE);
          ELSE
            W.ShowWindow(handle, W.SW_SHOW);
          END;
          property.AddBoolean("visible"  ,visible);
        END;

        IF 8 IN property.modified THEN
          IF ~ enabled THEN
            W.EnableWindow(handle,W.False);
          ELSE
            W.EnableWindow(handle,W.True);
          END;
          property.AddBoolean("enabled"  ,enabled);
        END;

        IF 9 IN property.modified THEN
          property.AddString("caption"  ,caption);
        END;

        IF 10 IN property.modified THEN
          property.AddBoolean("tabstop" ,tabstop);
        END;

        IF 11 IN property.modified THEN
          AlignControls;
          property.AddLongint("align" ,align);
        END;

        IF 12 IN property.modified THEN
          IF brush # 0 THEN W.DeleteObject(brush) END;
          brush := W.CreateSolidBrush(color);
          W.SetClassLong(handle,W.GCL_HBRBACKGROUND,brush);
          W.InvalidateRect(handle,NIL,W.True);
          property.AddLongint("color" ,color);
        END;

        IF {13,14,15,16,17} * property.modified # {} THEN
          IF 13 IN property.modified THEN
            font.SetName(font.name);
          END;
          IF 15 IN property.modified THEN
            font.SetSize(font.height);
          END;
          IF 16 IN property.modified THEN
            font.SetStyle(font.style);
          END;
          FontChange;
          W.InvalidateRect(handle,NIL,W.False);
          property.AddFont("font" ,font);
        END;

        property.modified := {};
      END;
    END CheckProperty;

    PROCEDURE PopupHide*(control: Wincontrol);
    VAR
      i   : Longint;
      hide: Boolean;
    BEGIN
      hide := FALSE;
      IF screen # NIL THEN
        FOR i := 0 TO screen.count -1 DO
          IF (screen.list[i] IS Popupmenu) & (screen.list[i](Popupmenu).control # control) & screen.list[i](Popupmenu).visible THEN
            IF (control = NIL) & ~ hide THEN
              screen.list[i](Popupmenu).Hidemain;
              hide := TRUE;
            END;
            screen.list[i](Popupmenu).Hide;
            W.ShowWindow(screen.list[i](Popupmenu).handle,W.SW_HIDE);
          END;
        END;
      END;
    END PopupHide;

    PROCEDURE ShowProperty*;
    VAR
      i,t,j,k : Longint;
      obj   : Persistent;
      f     : G.Font;
    BEGIN
      FOR i := 0 TO property.count - 1 DO
        Trace.String(property.list[i](Property).name); Trace.Char(9X); Trace.String(" := ");
        t := property.list[i](Property).type;
        CASE t OF
        |  typebyte    : Trace.Char(CHR(property.list[i](Property).byte));
        |  typechar    : Trace.Char(property.list[i](Property).char);
        |  typeboolean : Trace.Boolean(property.list[i](Property).boolean);
        |  typeshortint: Trace.Int(property.list[i](Property).shortint);
        |  typeinteger : Trace.Int(property.list[i](Property).integer);
        |  typelongint : Trace.Int(property.list[i](Property).longint);
        |  typereal    : ;
        |  typelongreal: ;
        |  typepointer : ;
        |  typeset     : Trace.Set(property.list[i](Property).set);
        |  typestring  : Trace.String(property.list[i](Property).string);
        |  typeobject  :
                         obj := property.list[i](Property).object(Persistent);
                         IF obj # NIL THEN
                           Trace.Ln;
                           FOR j := 0 TO obj.property.count - 1 DO
                             Trace.Char(9X); Trace.String(obj.property.list[j](Property).name);
                             Trace.Ln;
                           END;
                         END;
        |  typefont   :
                         f := property.list[i](Property).font(G.Font);
                         IF f # NIL THEN
                           Trace.Ln;
                           Trace.Char(9X); Trace.String("name := "); Trace.String(f.name); Trace.Ln;
                           Trace.Char(9X); Trace.String("color := "); Trace.Int(f.color);  Trace.Ln;
                           Trace.Char(9X); Trace.String("height := "); Trace.Int(f.height); Trace.Ln;
                           Trace.Char(9X); Trace.String("style := "); Trace.Set(f.style);  Trace.Ln;
                         END;
        ELSE
        END;
        Trace.Ln;
      END;
      Trace.Ln;
    END ShowProperty;

    PROCEDURE WindowProc*(VAR message: Message);
    VAR
      control: Wincontrol;
      p,m    : W.Point;
      x,y    : Longint;
      tme    : W.TrackMouseEventEx;
    BEGIN

      CASE message.msg OF
      |  W.WM_LBUTTONDOWN   : mousedown := TRUE; message.button := mbLeft;
      |  W.WM_LBUTTONUP     : message.button := mbLeft;
      |  W.WM_RBUTTONDOWN   : mousedown := TRUE; message.button := mbRight;
      |  W.WM_RBUTTONUP     : message.button := mbRight;
      |  W.WM_MBUTTONDOWN   : mousedown := TRUE; message.button := mbMiddle;
      |  W.WM_MBUTTONUP     : message.button := mbMiddle;
      |  W.WM_LBUTTONDBLCLK : message.button := mbLeft;
      |  W.WM_RBUTTONDBLCLK : message.button := mbRight;
      |  W.WM_MBUTTONDBLCLK : message.button := mbMiddle;
      ELSE
      END;

      CASE message.msg OF
      |  W.WM_LBUTTONDOWN   : WMMouseDown(message);
      |  W.WM_LBUTTONUP     : WMMouseUp(message);
      |  W.WM_RBUTTONDOWN   : WMMouseDown(message);
      |  W.WM_RBUTTONUP     : WMMouseUp(message);
      |  W.WM_MBUTTONDOWN   : WMMouseDown(message);
      |  W.WM_MBUTTONUP     : WMMouseUp(message);
      |  W.WM_LBUTTONDBLCLK : WMDblClick(message);
      |  W.WM_RBUTTONDBLCLK : WMDblClick(message);
      |  W.WM_MBUTTONDBLCLK : WMDblClick(message);
      |  W.WM_MOUSEWHEEL    : WMMouseWheel(message);
      |  W.WM_KEYDOWN       : WMKeyDown(message);
      |  W.WM_KEYUP         : WMKeyUp(message);
      |  W.WM_CHAR	    : WMChar(message);
      |  W.WM_SETFOCUS	    : WMSetFocus(message);
      |  W.WM_KILLFOCUS	    : WMKillFocus(message);
      |  W.WM_PAINT         :
                              IF message.dc = 0 THEN
                                W.DefWindowProc(handle, message.msg, message.wparam, message.lparam);
                                message.dc := 1;
                                WindowProc(message);
                              ELSE
                                WMPaint(message);
                                message.dc     := 0;
                                message.result := 0;
                              END;
      |  W.WM_ERASEBKGND    : WMEraseBkgnd(message);
                              IF message.result = 0 THEN
                                W.DefWindowProc(handle, message.msg, message.wparam, message.lparam);
                              END;
      |  W.WM_MOUSEMOVE     :
                              tme.cbSize := SIZE(W.TrackMouseEventEx);
                              tme.hwndTrack := handle;
                              tme.dwFlags := W.TME_LEAVE+W.TME_HOVER;
                              W.TrackMouseEvent(tme);
                              IF ~ mouseinside THEN
                                mouseinside := TRUE;
                                WMMouseEnter;
                              END;
                              WMMouseMove(message);
      |  W.WM_MOUSELEAVE    :
                              mouseinside := FALSE;
                              mousehover  := FALSE;
                              WMMouseLeave;
      |  W.WM_MOUSEHOVER    :
                              IF ~ mousehover THEN
                                mousehover := TRUE;
                                WMMouseHover;
                              END;
      |  W.WM_ACTIVATE      :
                              IF (U.LoWord(message.wparam) = W.WA_INACTIVE) THEN
                                WMInActivate(message);
                              ELSE
                                WMActivate(message);
                              END;
      |  W.WM_NCLBUTTONDOWN :
                              WMNCMouseDown(message);
                              message.result := W.DefWindowProc(handle, message.msg, message.wparam, message.lparam);
      |  W.WM_MOVE	    :
                              WMMove(message);
                              UpdateBounds;
      |  W.WM_SIZE	    :
                              WMSize(message);
                              UpdateBounds;
                              AlignControls;
      |  W.WM_COMMAND	    :
                              IF controls # NIL THEN
                                control := controls.list[message.wparam](Wincontrol);
                                IF control # NIL THEN
                                  control.WMCommand(message);
                                END;
                              END;
      |  W.WM_CLOSE         :
                              IF SELF = mainform THEN
                                W.PostQuitMessage(0);
                              ELSE
                                SELF.Hide;
                              END;
      |  W.WM_DESTROY       : IF brush # 0 THEN W.DeleteObject(brush) END;
                              IF font.handle # 0 THEN W.DeleteObject(font.handle) END;
      ELSE
        message.result := W.DefWindowProc(handle, message.msg, message.wparam, message.lparam);
      END;

      CASE message.msg OF
      |  W.WM_LBUTTONDOWN,
         W.WM_RBUTTONDOWN,
         W.WM_MBUTTONDOWN : W.SetCapture(handle);

      |  W.WM_LBUTTONUP,
         W.WM_RBUTTONUP,
         W.WM_MBUTTONUP   :
                            IF mousedown THEN
                              W.GetCursorPos(p);
                              W.ClientToScreen(handle,m);
                              IF (p.x>=m.x) & (p.y>=m.y) & (p.x<m.x+width) & (p.y<m.y+height) THEN
                                WMMouseClick(message);
                              END;
                            END;
                            W.ReleaseCapture();
                            mousedown := FALSE;
      |  W.WM_KEYDOWN,
         W.WM_KEYUP,
         W.WM_CHAR	    : ;
      ELSE
      END;

    END WindowProc;

  END Wincontrol;

  Customcontrol* = OBJECT (Wincontrol)
  VAR
    canvas*: G.Bitmap;

    PROCEDURE & InitControl*(Parent: Wincontrol);
    BEGIN
      Initctl^(Parent);
      classname    := "Customcontrol";
      width        := 250;
      height       := 250;
      NEW(canvas);
      canvas.SetSize(width,height);
    END InitControl;

    PROCEDURE Destroy*;
    BEGIN
      canvas.Free;
      Destroy^;
    END Destroy;

    PROCEDURE CreateParams*;
    BEGIN
      CreateParams^;
      style     := style - W.WS_BORDER;
      canvas.SetSize(width,height);
    END CreateParams;

    PROCEDURE WMSize*(VAR message: Message);
    BEGIN
      WMSize^(message);
      canvas.SetSize(width,height);
    END WMSize;

    PROCEDURE WMEraseBkgnd*(VAR message: Message);
    BEGIN
      message.result := 1;
    END WMEraseBkgnd;

    PROCEDURE Paint*;
    BEGIN
    END Paint;

    PROCEDURE WMPaint*(VAR message: Message);
    VAR
      dc : W.HDC;
    BEGIN
      dc := W.GetDC(handle);
      canvas.FillRect(0,0,width,height,color);
      Paint;
      canvas.Draw(dc,0,0);
      W.ReleaseDC(handle,dc);
    END WMPaint;

  END Customcontrol;

  Mainmenu* = OBJECT (Customcontrol)

    PROCEDURE & Create*(Parent: Wincontrol);
    BEGIN
      InitControl^(Parent);
      classname    := "Mainmenu";
    END Create;

  END Mainmenu;

  Popupmenu* = OBJECT (Customcontrol)
  VAR
    control*: Wincontrol;

    PROCEDURE & InitPopup*(AOwner: Wincontrol);
    BEGIN
      InitControl^(NIL);
      control        := AOwner;
      classname      := "Popupmenu";
      caption        := "";
      left           := 0;
      top            := 0;
      width          := 1;
      height         := 1;
      visible        := FALSE;
      color          := G.clBtnFace;
    END InitPopup;

    PROCEDURE Hidemain*;
    BEGIN
    END Hidemain;

  END Popupmenu;

  ObjectArray  = POINTER TO ARRAY OF Pointer;

  List* = OBJECT
  VAR
    list*: ObjectArray;
    count-: Longint;

    PROCEDURE & InitList*(size: Longint);
    BEGIN
      IF size <= 0 THEN size := 8 END;
      NEW(list,size);
      count := 0;
    END InitList;

    PROCEDURE Length*( ): Longint;
    BEGIN
      RETURN count
    END Length;

    PROCEDURE Grow;
    VAR
      old: ObjectArray;
      i: Longint;
    BEGIN
      old := list;
      NEW(list, (LEN(list) * 3+1) DIV 2);
      FOR i := 0 TO count - 1 DO list[i] := old[i] END
    END Grow;

    PROCEDURE Add*(x: Pointer);
    BEGIN
      IF count = LEN(list) THEN Grow END;
      list[count] := x;
      INC(count)
    END Add;

    PROCEDURE Remove*(x: Pointer);
    VAR
      i: Longint;
    BEGIN
      i := 0;
      WHILE (i < count) & (list[i] # x) DO INC( i ) END;
      IF i < count THEN
	WHILE (i < count - 1) DO
          list[i] := list[i + 1];
          INC( i )
        END;
	DEC(count);
        list[count] := NIL;
      END
    END Remove;

    PROCEDURE Insert*(i: Longint; x: Pointer);
    VAR
      j: Longint;
    BEGIN
      IF count = LEN(list) THEN Grow END;
      INC(count);
      j := count - 2;
      WHILE (j >= i) DO
        list[j+1] := list[j];
        DEC(j);
      END;
      list[i] := x;
    END Insert;

    PROCEDURE IndexOf*(x: Pointer): Longint;
    VAR
      i: Longint;
    BEGIN
      i := 0;
      WHILE i < count DO
        IF list[i] = x THEN RETURN i END;
	INC(i);
      END;
      RETURN -1
    END IndexOf;

    PROCEDURE SwapByIndex*(i,s: Longint);
    VAR
      x: Pointer;
    BEGIN
      IF (i >= 0) & (i < count) & (s >= 0) & (s < count) & (i # s) THEN
        x       := list[i];
        list[i] := list[s];
        list[s] := x;
      END
    END SwapByIndex;

    PROCEDURE Clear*;
    VAR
      i: Longint;
    BEGIN
      FOR i := 0 TO count - 1 DO list[i] := NIL END;
      count := 0;
    END Clear;

    PROCEDURE QuickSort(sortfunc: Sortfunc; lo, hi: Longint);
    VAR
      i, j: Longint;
      x, t: Pointer;
    BEGIN
      i := lo;
      j := hi;
      x := list[(lo + hi) DIV 2];
      WHILE i <= j DO
        WHILE sortfunc(list[i], x) DO INC(i) END;
        WHILE sortfunc(x, list[j]) DO DEC(j) END;
        IF i <= j THEN
	  t := list[i];
          list[i] := list[j];
          list[j] := t;
	  INC(i);
          DEC(j);
        END
      END;
      IF lo < j THEN QuickSort(sortfunc, lo, j) END;
      IF i < hi THEN QuickSort(sortfunc, i, hi) END
    END QuickSort;

    PROCEDURE SortName(n1,n2: Pointer): Boolean;
    BEGIN
      IF n1(Property).name < n2(Property).name THEN
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END SortName;

    PROCEDURE Sort*;
    BEGIN
      IF count > 0 THEN
        QuickSort(SortName, 0, count - 1);
      END;
    END Sort;

  END List;

  Propertylist* = OBJECT (List)
  VAR
    modified* : Set;
    property  : Property;

    PROCEDURE & Init*;
    BEGIN
      InitList^(0);
      modified := {};
    END Init;

    PROCEDURE AddByte*(name: String; value: Byte);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.byte := value;
      ELSE
        NEW(property,name,typebyte);
        property.byte := value;
        Add(property);
      END;
    END AddByte;

    PROCEDURE AddChar*(name: String; value: Char);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.char := value;
      ELSE
        NEW(property,name,typechar);
        property.char := value;
        Add(property);
      END;
    END AddChar;

    PROCEDURE AddBoolean*(name: String; value: Boolean);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.boolean := value;
      ELSE
        NEW(property,name,typeboolean);
        property.boolean := value;
        Add(property);
      END;
    END AddBoolean;

    PROCEDURE AddShortint*(name: String; value: Shortint);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.shortint := value;
      ELSE
        NEW(property,name,typeshortint);
        property.shortint := value;
        Add(property);
      END;
    END AddShortint;

    PROCEDURE AddInteger*(name: String; value: Integer);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.integer := value;
      ELSE
        NEW(property,name,typeinteger);
        property.integer := value;
        Add(property);
      END;
    END AddInteger;

    PROCEDURE AddLongint*(name: String; value: Longint);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.longint := value;
      ELSE
        NEW(property,name,typelongint);
        property.longint := value;
        Add(property);
      END;
    END AddLongint;

    PROCEDURE AddReal*(name: String; value: Real);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.real := value;
      ELSE
        NEW(property,name,typereal);
        property.real := value;
        Add(property);
      END;
    END AddReal;

    PROCEDURE AddLongreal*(name: String; value: Longreal);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.longreal := value;
      ELSE
        NEW(property,name,typelongreal);
        property.longreal := value;
        Add(property);
      END;
    END AddLongreal;

    PROCEDURE AddPointer*(name: String; value: Pointer);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.pointer := value;
      ELSE
        NEW(property,name,typepointer);
        property.pointer := value;
        Add(property);
      END;
    END AddPointer;

    PROCEDURE AddFont*(name: String; value: G.Font);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.font.name  := value.name;
        result.font.color := value.color;
        result.font.height := value.height;
        result.font.style := value.style;
      ELSE
        NEW(property,name,typefont);
        NEW(property.font,value.name);
        property.font.color := value.color;
        property.font.height := value.height;
        property.font.style := value.style;
        Add(property);
      END;
    END AddFont;

    PROCEDURE AddSet*(name: String; value: Set);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        result.set := value;
      ELSE
        NEW(property,name,typeset);
        property.set := value;
        Add(property);
      END;
    END AddSet;

    PROCEDURE AddString*(name: String; value: String);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        U.Copy(value,result.string);
      ELSE
        NEW(property,name,typestring);
        U.Copy(value,property.string);
        Add(property);
      END;
    END AddString;

    PROCEDURE AddObject*(name: String; value: Object);
    VAR
      result: Property;
    BEGIN
      result := GetProperty(name);
      IF result # NIL THEN
        property.object := value;
      ELSE
        NEW(property,name,typeobject);
        property.object := value;
        Add(property);
      END;
    END AddObject;

    PROCEDURE FindByName*(VAR name: String): Property;
    VAR
      i     : Longint;
      result: Property;
    BEGIN
      result := NIL;
      IF count > 0 THEN
        i := 0;
        WHILE (i < count) & (result = NIL) DO
          IF list[i](Property).name[0] = name[0] THEN
            IF list[i](Property).name = name THEN
              result := list[i](Property);
            END;
          END;
          INC(i);
        END;
      END;
      RETURN result;
    END FindByName;

    PROCEDURE GetType*(name: String): Shortint;
    VAR
      result: Property;
    BEGIN
      result := FindByName(name);
      IF result # NIL THEN
        RETURN result.type;
      ELSE
        RETURN -1;
      END;
    END GetType;

    PROCEDURE GetProperty*(name: String): Property;
    BEGIN
      RETURN FindByName(name);
    END GetProperty;

  END Propertylist;

  Classlist* = OBJECT
  VAR
    left   : Classlist;
    right  : Classlist;
    name   : Name;
    sys    : Boolean;
    count* : Longint;

    PROCEDURE & Init*;
    BEGIN
      left    := NIL;
      right   := NIL;
      name    := "";
      count   := 0;
    END Init;

    PROCEDURE Add*(name: String; sys: Boolean);
    VAR
      ob0, ob1: Classlist;
      left: Boolean;
      res: Name;
      str : ARRAY 255 OF Char;
    BEGIN

      ob0 := SELF;
      ob1 := ob0.right;
      left := FALSE;

      U.Uppercase(name, res);

      LOOP
        IF ob1 # NIL THEN
          IF res < ob1.name THEN
	    ob0 := ob1;
	    ob1 := ob0.left;
	    left := TRUE
          ELSIF res > ob1.name THEN
	    ob0 := ob1;
	    ob1 := ob0.right;
	    left := FALSE
          ELSE
            EXIT
          END;
        ELSE (*insert*)
          NEW(ob1);
          IF left THEN
	    ob0.left := ob1
          ELSE
	    ob0.right := ob1
          END;
          ob1.left:=NIL;
          ob1.right:=NIL;
          ob1.name := res;
          ob1.sys := sys;
          INC(count);
          EXIT;
        END;
      END;

    END Add;

    PROCEDURE Find*(name: String): Shortint;
    VAR
      obj: Classlist;
      res: Name;
    BEGIN
      U.Uppercase(name, res);
      obj := SELF;
      LOOP
        IF obj # NIL THEN
          IF res < obj.name THEN
	    obj := obj.left;
          ELSIF res > obj.name THEN
	    obj := obj.right;
          ELSE (*ok*)
            EXIT
          END;
        ELSE (*no found*)
          RETURN noClass;
        END;
      END;
      IF obj.sys THEN
        RETURN sysClass;
      ELSE
        RETURN userClass;
      END;
    END Find;

    PROCEDURE Delete*(name: String);
    VAR
      obj: Classlist;
      res: Name;
    BEGIN
      U.Uppercase(name, res);
      obj := SELF;
      LOOP
        IF obj # NIL THEN
          IF res < obj.name THEN
	    obj := obj.left;
          ELSIF res > obj.name THEN
	    obj := obj.right;
          ELSE (*ok*)
            obj.name := "";
            EXIT
          END;
        ELSE (*no found*)
          EXIT;
        END;
      END;
    END Delete;

    PROCEDURE UnregisterClass*;

      PROCEDURE Show(VAR obj: Classlist);
      BEGIN
        IF obj = NIL THEN RETURN END;
        IF obj.left  # NIL THEN Show(obj.left); END;
        IF obj.right # NIL THEN Show(obj.right); END;
        IF ~obj.sys THEN W.UnregisterClass(obj.name,W.hInstance); END;
      END Show;

    BEGIN
      Show(left);
      Show(right);
    END UnregisterClass;

  END Classlist;

  Data = RECORD
    address: Longint;
    length : Longint;
    size   : Longint;
  END;

  Stringlist* = OBJECT
  VAR
    heap     : Longint;
    info     : Longint;
    size     : Longint;
    sizedata : Longint;
    data     : Data;
    count-   : Longint;

    PROCEDURE & Create*;
    BEGIN
      sizedata := SIZE(Data);
      size     := sizedata * 2;
      heap     := W.HeapCreate({},size,0);
      info     := 0;
      count    := 0;
    END Create;

    PROCEDURE Free*;
    BEGIN
      IF heap > 0 THEN W.HeapDestroy(heap) END;
    END Free;

    PROCEDURE GetLineData(index: Longint);
    BEGIN
      SYSTEM.MOVE(info+index*sizedata,SYSTEM.ADR(data),sizedata);
    END GetLineData;

    PROCEDURE SetLineData(index: Longint);
    BEGIN
      SYSTEM.MOVE(SYSTEM.ADR(data),info+index*sizedata,sizedata);
    END SetLineData;

    PROCEDURE GrowLine(size,index: Longint);
    BEGIN
      GetLineData(index);
      size := (size*3+1) DIV 2;
      data.size    := size;
      data.address := W.HeapReAlloc(heap,W.HEAP_ZERO_MEMORY,data.address,size);
      SetLineData(index);
    END GrowLine;

    PROCEDURE ShortLine(index: Longint);
    VAR
      newadr: Longint;
    BEGIN
      GetLineData(index);
      newadr := W.HeapAlloc(heap,W.HEAP_ZERO_MEMORY,data.length+1);
      SYSTEM.MOVE(data.address,newadr,data.length);
      W.HeapFree(heap,{},data.address);
      data.size    := data.length+1;
      data.address := newadr;
      SetLineData(index);
    END ShortLine;

    PROCEDURE Grow;
    BEGIN
      size := ((count+1)*sizedata*3+1) DIV 2;
      info := W.HeapReAlloc(heap,W.HEAP_ZERO_MEMORY,info,size);
    END Grow;

    PROCEDURE Short;
    VAR
      newheap: Longint;
      newinfo: Longint;
      newaddr: Longint;
      i      : Longint;
    BEGIN
      IF heap > 0 THEN
        size     := count * sizedata;
        newheap  := W.HeapCreate({},size,0);
        IF newheap > 0 THEN
          newinfo  := W.HeapAlloc(newheap,W.HEAP_ZERO_MEMORY,size);
          FOR i := 0 TO count - 1 DO
            SYSTEM.MOVE(info+i*sizedata,SYSTEM.ADR(data),sizedata);
            newaddr := W.HeapAlloc(newheap,W.HEAP_ZERO_MEMORY,data.length+1);
            SYSTEM.MOVE(data.address,newaddr,data.length);
            data.address := newaddr;
            data.size    := data.length+1;
            SYSTEM.MOVE(SYSTEM.ADR(data),newinfo+i*sizedata,sizedata);
          END;
        END;
        W.HeapDestroy(heap);
        heap := newheap;
        info := newinfo;
      END;
    END Short;

    PROCEDURE Clear*;
    BEGIN
      IF heap > 0 THEN
        W.HeapDestroy(heap);
        size  := sizedata * 2;
        heap  := W.HeapCreate({},size,0);
        info  := 0;
      END;
      count := 0;
    END Clear;

    PROCEDURE ClearLine(start,end,index: Longint);
    VAR
      i: Longint;
    BEGIN
      GetLineData(index);
      IF data.size <= end + 1 THEN GrowLine(end + 1,index) END;
      FOR i := start TO end - 1 DO SYSTEM.PUT8(data.address+i,20X) END;
    END ClearLine;

    PROCEDURE InsertChar*(pos,index: Longint; c: Char);
    VAR
      i: Longint;
    BEGIN
      IF (heap > 0) & (info > 0) & (count > 0) & (index >= 0) & (index < count) & (pos >= 0) THEN
        GetLineData(index);
        IF pos <= data.length THEN
          IF data.size <= data.length + 2 THEN GrowLine(data.length + 2,index) END;
          FOR i := data.length TO pos BY -1 DO
            SYSTEM.PUT8(data.address+i+1,SYSTEM.GET8(data.address+i));
          END;
          INC(data.length);
        ELSIF pos > data.length THEN
          ClearLine(data.length,pos,index);
          data.length  := pos + 1;
        END;
        SYSTEM.PUT8(data.address+pos,c);
        SetLineData(index);
      END;
    END InsertChar;

    PROCEDURE DeleteChar*(pos,index: Longint): Longint;
    VAR
      res: Longint;
    BEGIN
      res := 0;
      IF (heap > 0) & (info > 0) & (count > 0) & (index >= 0) & (index < count) & (pos >= 0) THEN
        GetLineData(index);
        IF (data.length > 0) & (data.length < data.size DIV 2) THEN ShortLine(index) END;
        IF pos = data.length - 1 THEN
          SYSTEM.PUT8(data.address+pos,0X);
          DEC(data.length);
          res := data.length;
        ELSIF pos < data.length - 1 THEN
          SYSTEM.MOVE(data.address+pos+1,data.address+pos,data.length - pos - 1);
          SYSTEM.PUT8(data.address+data.length-1,0X);
          DEC(data.length);
          res := data.length;
        ELSIF pos > data.length - 1 THEN
          GrowLine(pos,index);
          ClearLine(data.length,pos,index);
          data.length := pos;
          res := 0;
        END;
        SetLineData(index);
      END;
      RETURN res;
    END DeleteChar;

    PROCEDURE Add*(s: String);
    BEGIN
      IF (heap > 0) THEN
        IF info = 0 THEN info := W.HeapAlloc(heap,W.HEAP_ZERO_MEMORY,size) END;
        IF (count+1)*sizedata >= size THEN Grow END;
        IF s # "" THEN
          data.length  := U.Length(s);
          data.size    := data.length+1;
          data.address := W.HeapAlloc(heap,W.HEAP_ZERO_MEMORY,data.length+1);
          SYSTEM.MOVE(SYSTEM.ADR(s[0]),data.address,data.length);
        ELSE
          data.length  := 0;
          data.size    := 1;
          data.address := W.HeapAlloc(heap,W.HEAP_ZERO_MEMORY,1);
        END;
        SetLineData(count);
        INC(count);
      END;
    END Add;

    PROCEDURE Insert*(index: Longint; s: String);
    VAR
      i: Longint;
    BEGIN
      IF (heap > 0) & (index >= 0) & (index < count) THEN
        IF info = 0 THEN info := W.HeapAlloc(heap,W.HEAP_ZERO_MEMORY,size) END;
        IF (count+2)*sizedata >= size THEN Grow END;
        FOR i := count-1 TO index BY -1 DO
          SYSTEM.MOVE(info+i*sizedata,info+(i+1)*sizedata,sizedata);
        END;
        IF s # "" THEN
          data.length  := U.Length(s);
          data.size    := data.length+1;
          data.address := W.HeapAlloc(heap,W.HEAP_ZERO_MEMORY,data.length+1);
          SYSTEM.MOVE(SYSTEM.ADR(s[0]),data.address,data.length);
        ELSE
          data.length  := 0;
          data.size    := 1;
          data.address := W.HeapAlloc(heap,W.HEAP_ZERO_MEMORY,1);
        END;
        SetLineData(index);
        INC(count);
      ELSIF count = 0 THEN
        Add(s);
      END;
    END Insert;

    PROCEDURE Delete*(index: Longint);
    VAR
      i: Longint;
    BEGIN
      IF (heap > 0) THEN
        IF (count > 1) & (index >= 0) & (index < count) THEN
          GetLineData(index);
          IF data.address > 0 THEN
            W.HeapFree(heap,{},data.address);
          END;
          FOR i := index + 1 TO count - 1  DO
            SYSTEM.MOVE(info+i*sizedata,info+(i-1)*sizedata,sizedata);
          END;
          DEC(count);
          IF count < (size DIV sizedata) DIV 2 THEN Short END;
        ELSIF (index = 0) & (count = 1) THEN
          Clear;
        END;
      END;
    END Delete;

    PROCEDURE SplitLine*(pos,index: Longint);
    VAR
      i        : Longint;
      startline: Longint;
      startlen : Longint;
      startadr : Longint;
      nextsize : Longint;
      nextline : Longint;
      nextlen  : Longint;
      nextadr  : Longint;
    BEGIN
      IF (heap > 0) & (info > 0) & (count > 0) THEN
        GetLineData(index);
        startline:= index;
        startlen := data.length;
        startadr := data.address;
        i        := data.length - pos;
        IF index = count -1 THEN
          Add("");
        ELSIF index < count - 1 THEN
          Insert(index+1,"");
        END;
        GetLineData(index+1);
        nextline := index+1;
        nextlen  := data.length;
        IF i > data.size - 2 THEN GrowLine(i + 2,nextline) END;
        nextsize := data.size;
        nextadr  := data.address;
        IF i > 0 THEN
          SYSTEM.MOVE(startadr+pos,nextadr,i);
          ClearLine(pos,startlen,startline);
          data.length := pos;
          SetLineData(startline);
        END;
        IF i < 0 THEN i := 0 END;
        data.address := nextadr;
        data.size := nextsize;
        data.length  := i;
        SetLineData(nextline);
      END;
    END SplitLine;

    PROCEDURE JoinLine*(pos,index: Longint);
    VAR
      i,j,a: Longint;
    BEGIN
      IF (heap > 0) & (info > 0) & (count > 1) & (index > 0) & (index <= count) THEN
        GetLineData(index);
        IF data.length > 0 THEN
          a := data.address;
          i := data.length;
          GetLineData(index-1);
          IF pos < 0 THEN
            j := data.length;
          ELSE
            j := pos;
          END;
          IF i+j+1 >= data.size - 2 THEN GrowLine(i + j + 1,index-1) END;
          SYSTEM.MOVE(a,data.address+j,i);
          data.length := j;
          INC(data.length,i);
          SetLineData(index-1);
        END;
        Delete(index);
      END;
    END JoinLine;

    PROCEDURE DeleteRange*(start,end: W.Point);
    VAR
      startline: Longint;
      startlen : Longint;
      startadr : Longint;
      startsize: Longint;
      endline  : Longint;
      endlen   : Longint;
      endadr   : Longint;
      i        : Longint;
    BEGIN
      IF (heap > 0) & (info > 0) & (count > 0) THEN

        IF start.y > end.y THEN
          U.Swap(start.y,end.y);
          U.Swap(start.x,end.x);
        ELSIF start.y = end.y THEN
          IF start.x > end.x THEN U.Swap(start.x,end.x) END;
        END;

        IF (start.x >= 0) & (start.y >= 0) & (end.y <= count) THEN

          GetLineData(start.y);
          startline := start.y;
          startlen := data.length;
          startadr := data.address;
          startsize:= data.size;

          GetLineData(end.y);
          endline  := end.y;
          endlen   := data.length;
          endadr   := data.address;
          IF end.x > data.length - 1 THEN end.x := data.length END;

          IF (start.x = 0) & (start.y = 0) & (end.x >= endlen) & (end.y = count - 1) THEN
            Clear;
            Add("");
          ELSIF (start.y = end.y) THEN
            SYSTEM.MOVE(startadr+end.x,startadr+start.x,startlen-end.x);
            ClearLine(start.x+startlen-end.x,startlen,startline);
            data.length := start.x+startlen-end.x;
            SYSTEM.PUT8(startadr+data.length,0X);
            SetLineData(startline);
          ELSE
            IF start.x + endlen - end.x >= startsize - 2 THEN
              GrowLine(start.x + endlen - end.x,startline);
              startadr := data.address;
            END;
            IF start.x > startlen -1 THEN
              ClearLine(startlen,start.x+endlen-end.x,startline);
            ELSE
              ClearLine(start.x,start.x+endlen-end.x,startline);
            END;
            SYSTEM.MOVE(endadr+end.x,startadr+start.x,endlen-end.x+1);
            data.length := start.x+endlen-end.x;
            SetLineData(startline);
            FOR i := endline TO startline + 1 BY -1 DO Delete(i) END;
          END;

        END;

      END;
    END DeleteRange;

    PROCEDURE CopyToMemory*(start,end: W.Point; VAR buffer: Longint);
    VAR
      startline: Longint;
      startlen : Longint;
      startadr : Longint;
      startsize: Longint;
      endline  : Longint;
      endlen   : Longint;
      endadr   : Longint;
      address  : Longint;
      i,sz     : Longint;

      PROCEDURE ClearLine(start,end: Longint);
      VAR
        i: Longint;
      BEGIN
        FOR i := start TO end DO SYSTEM.PUT8(address,20X); INC(address); END;
      END ClearLine;

    BEGIN
      IF (heap > 0) & (info > 0) & (count > 0) THEN

        IF start.y > end.y THEN
          U.Swap(start.y,end.y);
          U.Swap(start.x,end.x);
        ELSIF start.y = end.y THEN
          IF start.x > end.x THEN U.Swap(start.x,end.x) END;
        END;

        IF (start.x >= 0) & (start.y >= 0) & (end.y <= count) THEN

          GetLineData(start.y);
          startline := start.y;
          startlen := data.length;
          startadr := data.address;
          startsize:= data.size;

          GetLineData(end.y);
          endline  := end.y;
          endlen   := data.length;
          endadr   := data.address;

          sz       := 0;

          IF (start.y = end.y) THEN
            buffer := W.GlobalAlloc(W.GHND,end.x - start.x + 1);
            address  := W.GlobalLock(buffer);
            IF (start.x <= startlen) & (end.x <= endlen) THEN
              SYSTEM.MOVE(startadr+start.x,address,end.x-start.x);
            ELSIF (start.x < startlen) & (end.x > endlen) THEN
              SYSTEM.MOVE(startadr+start.x,address,startlen-start.x);
              INC(address,startlen-start.x);
              ClearLine(startlen,end.x - 1);
            ELSE
              ClearLine(start.x,end.x - 1);
            END;
            W.GlobalUnlock(buffer);
          ELSE

            FOR i := startline TO endline DO
              GetLineData(i);
              sz := sz + data.length + 4;
            END;
            IF end.x > endlen - 1 THEN
              sz := sz + end.x - endlen + 1;
            END;

            buffer := W.GlobalAlloc(W.GHND,sz);
            address  := W.GlobalLock(buffer);

            IF start.x <= startlen - 1 THEN
              SYSTEM.MOVE(startadr+start.x,address,startlen-start.x);
              INC(address,startlen-start.x);
            END;
            SYSTEM.PUT16(address,0A0DH); INC(address,2);

            FOR i := startline + 1 TO endline - 1 DO
              GetLineData(i);
              IF data.length > 0 THEN
                SYSTEM.MOVE(data.address,address,data.length);
                INC(address,data.length);
              END;
              SYSTEM.PUT16(address,0A0DH); INC(address,2);
            END;

            IF end.x <= endlen - 1 THEN
              IF end.x > 0 THEN
                SYSTEM.MOVE(endadr,address,end.x);
                INC(address,end.x);
              ELSE
                SYSTEM.PUT16(address,0A0DH);
                INC(address,2);
              END;
            ELSE
              SYSTEM.MOVE(endadr,address,endlen);
              INC(address,endlen);
              ClearLine(endlen,end.x - 1);
            END;

            W.GlobalUnlock(buffer);
          END;

        END;

      END;
    END CopyToMemory;

    PROCEDURE PasteFromoMemory*(pos,index: Longint): W.Point;
    VAR
      buffer   : Longint;
      address  : Longint;
      datasize : Longint;
      i,cnt,j  : Longint;
      c        : Char;
      res      : W.Point;
    BEGIN
      res.x := pos;
      res.y := index;
      IF (heap > 0) & (info > 0) & (count > 0) THEN
        W.OpenClipboard(0);
        buffer := W.GetClipboardData(W.CF_TEXT);
        IF buffer # 0 THEN
          address  := W.GlobalLock(buffer);
          datasize := W.GlobalSize(buffer);
          i   := 0;
          cnt := 0;
          j   := pos;
          c := CHR(SYSTEM.GET8(address));
          WHILE (i < datasize-1) & (c # 0X) DO
            WHILE (i < datasize) & (c = 0DX) DO
              INC(i); c := CHR(SYSTEM.GET8(address+i));
              IF (c = 0AX) THEN INC(i); c := CHR(SYSTEM.GET8(address+i)); END;
              SplitLine(j,index+cnt);
              j := 0;
              INC(cnt);
            END;
            IF c # 0X THEN
              CASE c OF
              |  09X :
                           InsertChar(j,index+cnt,20X); INC(j);
                           InsertChar(j,index+cnt,20X); INC(j);
              |  01X..08X,
                 0AX..1FX: ;
              ELSE
                InsertChar(j,index+cnt,c); INC(j);
              END;
            END;
            INC(i);
            c := CHR(SYSTEM.GET8(address+i));
          END;
          W.GlobalUnlock(buffer);
        END;
        W.CloseClipboard();
        res.x := j;
        res.y := index+cnt;
      END;
      RETURN res;
    END PasteFromoMemory;

    PROCEDURE Length*(index: Longint): Longint;
    BEGIN
      IF (heap > 0) & (info > 0) &  (count > 0) & (index >= 0) & (index < count) THEN
        GetLineData(index);
        RETURN data.length;
      ELSE
        RETURN 0;
      END;
    END Length;

    PROCEDURE CharByPos*(pos,index: Longint): Char;
    VAR
      c: Char;
    BEGIN
      c := 20X;
      IF (heap > 0) & (info > 0) & (index >= 0) & (index < count) THEN
        GetLineData(index);
        IF pos < data.length THEN c := CHR(SYSTEM.GET8(data.address+pos)) END;
      END;
      RETURN c;
    END CharByPos;

    PROCEDURE Strings*(index: Longint): String;
    VAR
      s: ARRAY 4096 OF Char;
    BEGIN
      IF (heap > 0) & (info > 0) & (index >= 0) & (index <= count - 1) THEN
        GetLineData(index);
        IF data.length > 0 THEN
          IF data.length >= 4096 THEN data.length := 4096 - 1 END;
          SYSTEM.MOVE(data.address,SYSTEM.ADR(s[0]),data.length);
          s[data.length+1] := 0X;
        ELSE
          s[0] := 0X;
        END;
      ELSE
        s[0] := 0X;
      END;
      RETURN s;
    END Strings;

    PROCEDURE SubString*(pos1,pos2,index: Longint): String;
    VAR
      s     : ARRAY 4096 OF Char;
      length: Longint;
    BEGIN
      IF (heap > 0) & (info > 0) & (index >= 0) & (index < count) THEN
        GetLineData(index);
        IF pos2 > data.length THEN pos2 := data.length END;
        IF pos2 < 0 THEN pos2 := data.length END;
        length := (pos2-pos1)+1;
        IF pos2 >= pos1 THEN
          IF  length >= 4096 THEN length := 4096 - 1 END;
          SYSTEM.MOVE(data.address+pos1,SYSTEM.ADR(s[0]),length);
          s[length] := 0X;
        ELSE
          s[0] := 0X;
        END;
      ELSE
        s[0] := 0X;
      END;
      RETURN s;
    END SubString;

    PROCEDURE LoadFromFile*(name: String);
    VAR
      file    : F.File;
      i,j     : Longint;
      filelen : Longint;
      address : Longint;
      c       : Char;
      s       : ARRAY 4096 OF Char;
    BEGIN
      Clear;

      IF heap > 0 THEN

        filelen := 0;
        address  := 0;
        NEW(file);
        IF file.Open(name) THEN
          filelen := file.Length()+1;
          address := W.VirtualAlloc(filelen);
          IF address > 0 THEN file.ReadMemory(address,filelen-1) END;
        END;
        file.Close;
        file := NIL;

        IF filelen > 0 THEN

          i := 0;
          j := 0;

          IF info = 0 THEN info := W.HeapAlloc(heap,W.HEAP_ZERO_MEMORY,size) END;

          WHILE j < filelen DO

            c := CHR(SYSTEM.GET8(address+j));

            IF i <= LEN(s)-1 THEN

              WHILE (j < filelen) & (c = 0DX) DO
                INC(j); c := CHR(SYSTEM.GET8(address+j));
                IF (c = 0AX) THEN INC(j); c := CHR(SYSTEM.GET8(address+j)); END;
                s[i] := 0X;
                i := 0;
                Add(s);
              END;

              WHILE (j < filelen) & (c < 20X) & (c # 09X) DO
                INC(j); c := CHR(SYSTEM.GET8(address+j));
                (*
                CASE c OF
                  0X..1FX: s[i] := " ";
                ELSE
                END;
                INC(i);
                *)
              END;

              WHILE (j < filelen) & (c = 09X) DO
                INC(j); c := CHR(SYSTEM.GET8(address+j));
                s[i] := 20X; INC(i);
                s[i] := 20X; INC(i);
              END;

              IF (j < filelen) THEN s[i] := c; INC(i); END;

            END;

            INC(j);

          END;

          IF i > 0 THEN s[i] := 0X; Add(s); END;

        END;

        IF address > 0 THEN W.VirtualFree(address) END;

      END;

    END LoadFromFile;

  END Stringlist;

PROCEDURE [WINAPI] WndProc(hwnd: W.HWND; uMsg: Longint; wParam: W.WParam; lParam: W.LParam): W.LResult;
VAR
  control: Wincontrol;
  message: Message;
BEGIN

  message.msg    := uMsg;
  message.wparam := wParam;
  message.lparam := lParam;
  message.result := 0;
  message.button := 0;
  message.dc     := 0;

  control := SYSTEM.VAL(Wincontrol,W.GetProp(hwnd,SYSTEM.ADR("Wincontrol")));
  IF control # NIL THEN
    control.WindowProc(message);
  ELSE
    RETURN W.DefWindowProc(hwnd, uMsg, wParam, lParam);
  END;
  RETURN message.result;

END WndProc;

PROCEDURE [WINAPI] TimerProc(hwnd: W.HWND; uMsg, idEvent, dwTime: Longint): Longint;
VAR
  i: Longint;
BEGIN
  IF (screen # NIL) & (screen.count > 0) THEN
    FOR i := 0 TO screen.count-1 DO
      IF (screen.list[i] IS Timer) & (screen.list[i](Timer).timerid = idEvent) THEN
        screen.list[i](Timer).Timer;
        RETURN 0;
      END;
    END;
  END;
  RETURN 0;
END TimerProc;

PROCEDURE Run*;
VAR
  msg    : W.Msg;
  i      : Longint;
BEGIN
  FOR i:=0 TO screen.count - 1 DO
    IF ~ (screen.list[i] IS Timer) THEN
      screen.list[i](Wincontrol).CreateControl;
    END;
  END;
  WHILE (W.GetMessage(msg,0,0,0) # 0) DO
    W.TranslateMessage(msg);
    W.DispatchMessage(msg);
    FOR i:=0 TO screen.count - 1 DO
      IF ~ (screen.list[i] IS Timer) & screen.list[i](Wincontrol).enabled THEN
        screen.list[i](Wincontrol).Update;
        screen.list[i](Wincontrol).CheckProperty;
      END;
    END;
  END;
  FOR i:=0 TO screen.count - 1 DO
    IF ~ (screen.list[i] IS Timer) THEN
      screen.list[i](Wincontrol).Destroy;
      W.DestroyWindow(screen.list[i](Wincontrol).handle);
    ELSE
      screen.list[i](Timer).Free;
    END;
  END;
  screen.Clear;
  screen := NIL;
  classlist.UnregisterClass;
  W.ExitProcess(0);
END Run;

BEGIN
  NEW(classlist);

  classlist.Add("Button"	    ,TRUE);
  classlist.Add("ComboBox"	    ,TRUE);
  classlist.Add("Edit"		    ,TRUE);
  classlist.Add("ListBox"	    ,TRUE);
  classlist.Add("MDIClient"	    ,TRUE);
  classlist.Add("ScrollBar"	    ,TRUE);
  classlist.Add("Static"	    ,TRUE);

  classlist.Add("ToolbarWindow32"   ,TRUE);
  classlist.Add("ReBarWindow32"	    ,TRUE);
  classlist.Add("ComboBoxEx32"	    ,TRUE);
  classlist.Add("tooltips_class32"  ,TRUE);

  classlist.Add("SysListView32"	    ,TRUE);
  classlist.Add("SysTreeView32"	    ,TRUE);
  classlist.Add("SysHeader32"	    ,TRUE);
  classlist.Add("SysTabControl32"   ,TRUE);
  classlist.Add("SysAnimate32"	    ,TRUE);
  classlist.Add("SysMonthCal32"	    ,TRUE);
  classlist.Add("SysDateTimePick32" ,TRUE);
  classlist.Add("SysIPAddress32"    ,TRUE);

  classlist.Add("msctls_statusbar32",TRUE);
  classlist.Add("msctls_trackbar32" ,TRUE);
  classlist.Add("msctls_updown32"   ,TRUE);
  classlist.Add("msctls_progress32" ,TRUE);
  classlist.Add("msctls_hotkey32"   ,TRUE);
END Classes.
