MODULE Dialogs;

IMPORT
  SYSTEM, Kernel, W := Windows, C := Controls;

TYPE
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

  Opendialog* = OBJECT
  VAR
    filename*    : ARRAY 256 OF Char;
    filter*      : ARRAY 256 OF Char;
    title*       : ARRAY 256 OF Char;
    ext*         : ARRAY 10  OF Char;
    filterindex* : Longint;

    owner        : C.Control;
    dialog       : W.OpenFileNameA;

    PROCEDURE & Create*(Parent: C.Control);
    BEGIN
      filename[0] := 0X;
      filterindex := 1;
      IF Parent # NIL THEN
        owner := Parent;
      END;
      title[0] := 0X;
    END Create;

    PROCEDURE Convert (VAR a: String);
    VAR
      i: Longint;
    BEGIN
      i := 0;
      WHILE a[i] # 0X DO
        IF a[i]='|' THEN a[i]:=0X END;
        INC(i);
      END;
    END Convert;

    PROCEDURE FillRec();
    BEGIN
      dialog.lStructSize       := SIZE(W.OpenFileNameA);
      dialog.hwndOwner         := owner.handle;
      dialog.hInstance         := W.hInstance;
      dialog.lpstrFilter       := SYSTEM.ADR(filter);
      dialog.lpstrCustomFilter := 0;
      dialog.nMaxCustFilter    := 0;
      dialog.nFilterIndex      := filterindex;
      dialog.lpstrFile         := SYSTEM.ADR(filename);
      dialog.nMaxFile          := LEN(filename);
      dialog.lpstrFileTitle    := W.NULL;
      dialog.nMaxFileTitle     := 0;
      dialog.lpstrInitialDir   := W.NULL;
      IF title[0] = 0X THEN
        dialog.lpstrTitle        := W.NULL;
      ELSE
        dialog.lpstrTitle        := SYSTEM.ADR(title);
      END;
      dialog.Flags             := W.OFN_ALLOWMULTISELECT + W.OFN_EXPLORER + W.OFN_PATHMUSTEXIST + W.OFN_FILEMUSTEXIST;
      dialog.nFileOffset       := 0;
      dialog.nFileExtension    := 0;
      dialog.lpstrDefExt       := SYSTEM.ADR(ext);
      dialog.lCustData         := 0;
      dialog.lpfnHook          := NIL;
      dialog.lpTemplateName    := W.NULL;

    END FillRec;

    PROCEDURE Execute* (): Boolean;
    BEGIN
      Convert(filter);
      FillRec();
      IF W.GetOpenFileName(dialog) # 0 THEN
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END Opendialog;

  Savedialog* = OBJECT (Opendialog)

    PROCEDURE Execute*(): Boolean;
    BEGIN
      Convert(filter);
      FillRec^();
      IF W.GetSaveFileName(dialog) # 0 THEN
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END Savedialog;

  Colordialog* = OBJECT
  VAR
    color*    : Longint;
    dialog    : W.ChooseColorA;
    acrCustClr: ARRAY 16 OF W.ColorRef;
    owner     : C.Control;

    PROCEDURE & Create*(Parent: C.Control);
    BEGIN
      IF Parent # NIL THEN
        owner := Parent;
      END;
      dialog.lStructSize     := SIZE(W.ChooseColorA);
      dialog.hwndOwner       := owner.handle;
      dialog.hInstance       := W.hInstance;
      dialog.rgbResult       := color;
      dialog.lpCustColors    := SYSTEM.ADR(acrCustClr);
      dialog.Flags           := W.CC_FULLOPEN + W.CC_RGBINIT;
      dialog.lCustData       := 0;
      dialog.lpfnHook        := NIL;
      dialog.lpTemplateName  := W.NULL;
    END Create;

    PROCEDURE Execute* (): BOOLEAN;
    BEGIN
      IF W.ChooseColor(dialog) # 0 THEN
        color := dialog.rgbResult;
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END Colordialog;

  Fontdialog* = OBJECT
  VAR
    font*      : W.HFont;
    fontname*  : ARRAY 32 OF CHAR;
    color*     : LONGINT;

    dialog     : W.ChooseFontA;
    logfont    : W.LogFontA;
    owner      : C.Control;

    PROCEDURE & Create*(Parent: C.Control);
    BEGIN
      IF Parent # NIL THEN
        owner := Parent;
      END;
      dialog.lStructSize := SIZE(W.ChooseFontA);
      dialog.hwndOwner   := owner.handle;
      dialog.hInstance   := W.hInstance;
      dialog.lpLogFont   := SYSTEM.ADR(logfont);
      dialog.rgbColors   := color;
      dialog.Flags       := W.CF_SCREENFONTS + W.CF_EFFECTS;
      dialog.lCustData   := 0;
    END Create;

    PROCEDURE Execute* (): BOOLEAN;
    BEGIN
      IF W.ChooseFont(dialog)# 0 THEN
        font := W.CreateFontIndirect(logfont);
        SYSTEM.MOVE(SYSTEM.ADR(logfont.lfFaceName),SYSTEM.ADR(fontname),LEN(logfont.lfFaceName));
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END Fontdialog;

  Printdialog* = OBJECT
  VAR
    dialog: W.PrintDlgA;
    owner : C.Control;

    PROCEDURE & Create*(Parent: C.Control);
    BEGIN
      IF Parent # NIL THEN
        owner := Parent;
      END;
      dialog.lStructSize        := SIZE(W.PrintDlgA)-2;
      dialog.hwndOwner          := owner.handle;
      dialog.hInstance          := W.hInstance;
      dialog.hDevMode           := 0;
      dialog.hDevNames          := 0;
      dialog.hDC                := 0;
      dialog.Flags              := W.PD_USEDEVMODECOPIESANDCOLLATE + W.PD_RETURNDC;
      dialog.nFromPage          := -1;
      dialog.nToPage            := -1;
      dialog.nMinPage           :=  1;
      dialog.nMaxPage           := -1;
      dialog.nCopies            :=  1;
      dialog.hInstance          :=  0;
      dialog.lCustData          := 0;
      dialog.lpfnPrintHook      := NIL;
      dialog.lpfnSetupHook      := NIL;
      dialog.lpPrintTemplateName:= W.NULL;
      dialog.lpSetupTemplateName:= W.NULL;
      dialog.hPrintTemplate     := 0;
      dialog.hSetupTemplate     := 0;

    END Create;

    PROCEDURE Execute* (): BOOLEAN;
    BEGIN
      IF W.PrintDlg(dialog) # 0 THEN
        W.DeleteDC(dialog.hDC);
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Execute;

  END Printdialog;

END Dialogs.

