MODULE Editors;

IMPORT
  SYSTEM, Trace, W := Windows, U := Utilites, Cl := Classes,
  C := Controls, G := Graphics, S := Stdctrls;

TYPE
  Byte         = W.Byte;
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

  Charwidths   = ARRAY 4 OF ARRAY 256 OF Longint;
  Select = RECORD
    start: W.Point;
    end  : W.Point;
    count: Longint;
    color* : Longint;
    font*  : G.StdPack;
  END;

  Customedit* = OBJECT (C.Control)
  VAR
    lines*       : Cl.Stringlist;
    canvas*      : G.Bitmap;
    caret*       : W.Point;
    select*      : Select;
    vscroll*     : S.Scrollbar;
    multiline*   : Boolean;

    caretvisible : Boolean;
    crdefault    : W.HCursor;
    crbeam       : W.HCursor;

    offset       : Longint;
    charwidths   : Charwidths;
    charheight   : Longint;
    lineheight   : Longint;
    caretheight  : Longint;
    caretoffset  : Longint;
    startline    : Longint;
    endline      : Longint;
    currentline  : Longint;
    mousepress   : Boolean;
    visiblerows  : Longint;

    PROCEDURE & Create*(Parent: C.Control);
    BEGIN
      Create^(Parent);
      classname      := "Customedit";
      width          := 350;
      height         := 250;
      color          := G.clWhite;

      caret.x        := 0;
      caret.y        := 0;
      select.start.x := -1;
      select.start.y := -1;
      select.end.x   := -1;
      select.end.y   := -1;
      select.color   := 00A56D53H;
      select.font.color := G.clWhite;

      caretvisible   := FALSE;
      mousepress     := FALSE;
      multiline      := FALSE;

      offset         := 0;
      caretheight    := 0;
      caretoffset    := 0;
      startline      := 0;
      endline        := 0;
      currentline    := 0;
      visiblerows    := 0;

      NEW(canvas);
      NEW(lines);
      lines.Add("");

    END Create;

    PROCEDURE Destroy*;
    BEGIN
      canvas.Free;
      canvas := NIL;
      lines.Free;
      lines   := NIL;
      Destroy^;
    END Destroy;

    PROCEDURE CreateParams*;
    BEGIN
      CreateParams^;
      style       := style - W.WS_BORDER;
      wndstyle    := wndstyle + W.CS_VREDRAW + W.CS_HREDRAW;
      crdefault   := W.LoadCursor(0,SYSTEM.VAL(W.LPSTR,W.IDC_IBEAM));
      crbeam      := W.LoadCursor(0,SYSTEM.VAL(W.LPSTR,W.IDC_IBEAM));
      canvas.SetSize(width,height);

      IF multiline THEN
        NEW(vscroll,SELF);
        vscroll.kind  := S.sbVertical;
        vscroll.align := Cl.alRight;
        vscroll.OnChange := ScrollChange;
      END;

      Calculate;
    END CreateParams;

    PROCEDURE ScrollChange*(Sender: Cl.Object);
    BEGIN
      IF multiline THEN
        startline := vscroll.position;
      END;
      W.InvalidateRect(handle,NIL,W.False);
    END ScrollChange;

    PROCEDURE FontChange*;
    BEGIN
      FontChange^;
      Calculate;
    END FontChange;

    PROCEDURE Calculate*;
    VAR
      fst  : Set;
      size : Longint;
      dc,i : Longint;
      f    : G.Font;
    BEGIN

      HideCaret;
      W.DestroyCaret();

      f   := font;
      fst := font.style;

      dc := W.GetDC(handle);
      font.SetStyle({});
      W.SelectObject(dc, font.handle);
      FOR i := 32 TO 256 - 1 DO
        W.GetCharWidth(dc,i,i,size);
        charwidths[0,i] := size;
      END;
      font.SetStyle({G.fsBold});
      W.SelectObject(dc, font.handle);
      FOR i := 32 TO 256 - 1 DO
        W.GetCharWidth(dc,i,i,size);
        charwidths[1,i] := size;
      END;
      font.SetStyle({G.fsItalic});
      W.SelectObject(dc, font.handle);
      FOR i := 32 TO 256 - 1 DO
        W.GetCharWidth(dc,i,i,size);
        charwidths[2,i] := size;
      END;
      font.SetStyle({G.fsBold,G.fsItalic});
      W.SelectObject(dc, font.handle);
      FOR i := 32 TO 256 - 1 DO
        W.GetCharWidth(dc,i,i,size);
        charwidths[3,i] := size;
      END;
      size := charwidths[0,32];
      FOR i := 0 TO 31 DO
        charwidths[0,i] := size;
        charwidths[1,i] := size;
        charwidths[2,i] := size;
        charwidths[3,i] := size;
      END;
      W.ReleaseDC(handle,dc);

      font := f;
      font.SetStyle(fst);

      canvas.font := font;
      charheight  := canvas.font.height+3;
      lineheight  := charheight + 4;
      caretheight := lineheight - 4;
      caretoffset := (lineheight - caretheight) DIV 2;

      W.CreateCaret(handle,0,2,lineheight - 4);

      ShowCaret;

      IF multiline THEN
        vscroll.min      := 0;
        vscroll.max      := lines.count - 1;
        vscroll.pagesize := visiblerows;
        vscroll.position := startline;
      END;
    END Calculate;

    PROCEDURE FontStyleToInt*(font: G.Font): Shortint;
    VAR
      res: Shortint;
    BEGIN
      res := 0;
      IF    font.style = {G.fsBold,G.fsItalic} THEN res := 3
      ELSIF font.style = {G.fsItalic}          THEN res := 2
      ELSIF font.style = {G.fsBold}            THEN res := 1 END;
      RETURN res;
    END FontStyleToInt;

    PROCEDURE CopyToClipboard*;
    VAR
      buffer: Longint;
    BEGIN
      IF select.count > 0 THEN
        lines.CopyToMemory(select.start,select.end,buffer);
        IF W.OpenClipboard(handle) # 0 THEN
          W.EmptyClipboard();
          W.SetClipboardData(W.CF_TEXT,buffer);
          W.CloseClipboard();
        END;
      END;
    END CopyToClipboard;

    PROCEDURE CutToClipboard*;
    VAR
      start : W.Point;
      end   : W.Point;
    BEGIN
      IF select.count > 0 THEN

        start := select.start;
        end   := select.end;

        IF start.y > end.y THEN
          U.Swap(start.y,end.y);
          U.Swap(start.x,end.x);
        ELSIF start.y = end.y THEN
          IF start.x > end.x THEN U.Swap(start.x,end.x) END;
        END;

        CopyToClipboard;
        lines.DeleteRange(select.start,select.end);
        select.count := 0;
        caret.x := start.x;
        caret.y := start.y;
        W.InvalidateRect(handle,NIL,W.False);

      END;
    END CutToClipboard;

    PROCEDURE PasteFromClipboard*;
    VAR
      start,end: W.Point;
    BEGIN

      IF select.count > 0 THEN

        start := select.start;
        end   := select.end;

        IF start.y > end.y THEN
          U.Swap(start.y,end.y);
          U.Swap(start.x,end.x);
        ELSIF start.y = end.y THEN
          IF start.x > end.x THEN U.Swap(start.x,end.x) END;
        END;

        lines.DeleteRange(start,end);
        select.count := 0;
        caret.x := start.x;
        caret.y := start.y;

      END;

      caret := lines.PasteFromoMemory(caret.x,caret.y);
      IF multiline THEN
        vscroll.max      := lines.count - 1;
      END;
      W.InvalidateRect(handle,NIL,W.False);

    END PasteFromClipboard;

    PROCEDURE Delete*;
    VAR
      start : W.Point;
      end   : W.Point;
    BEGIN
      IF select.count > 0 THEN

        start := select.start;
        end   := select.end;

        IF start.y > end.y THEN
          U.Swap(start.y,end.y);
          U.Swap(start.x,end.x);
        ELSIF start.y = end.y THEN
          IF start.x > end.x THEN U.Swap(start.x,end.x) END;
        END;

        lines.DeleteRange(select.start,select.end);
        select.count := 0;
        caret.x := start.x;
        caret.y := start.y;
        W.InvalidateRect(handle,NIL,W.False);

      END;
    END Delete;

    PROCEDURE WMSetFocus*(VAR message: Cl.Message);
    BEGIN
      WMSetFocus^(message);
      W.CreateCaret(handle,0,2,lineheight - 4);
      ShowCaret;
    END WMSetFocus;

    PROCEDURE WMKillFocus*(VAR message: Cl.Message);
    BEGIN
      WMKillFocus^(message);
      HideCaret;
      W.DestroyCaret();
      Cl.activeform := handle;
    END WMKillFocus;

    PROCEDURE MouseLeave*;
    BEGIN
      MouseLeave^;
      W.SetCursor(crdefault);
    END MouseLeave;

    PROCEDURE MouseEnter*;
    BEGIN
      MouseEnter^;
      W.SetClassLong(handle,W.GCL_HCURSOR,crbeam);
    END MouseEnter;

    PROCEDURE MouseDown*(shift: Set; button,x,y: Longint);
    BEGIN
      MouseDown^(shift,button,x,y);
      mousepress := button = Cl.mbLeft;
      W.SetFocus(handle);
      IF mousepress THEN
        CoordToCaret(x,y);
        select.start.x := caret.x;
        select.start.y := caret.y;
        select.end     := select.start;
        select.count   := 0;
        W.InvalidateRect(handle,NIL,W.False);
      END;
    END MouseDown;

    PROCEDURE MouseMove*(shift: Set; x,y: Longint);
    BEGIN
      MouseMove^(shift,x,y);
      IF mousepress THEN
        CoordToCaret(x,y);
        select.end.x := caret.x;
        select.end.y := caret.y;
        IF ~ ((select.start.x = select.end.x) & (select.start.y = select.end.y)) THEN
          select.count := ABS(select.start.y - select.end.y) + 1;
        ELSE
          select.count := 0;
        END;

        IF multiline THEN
          IF select.count > 0 THEN
            IF (startline > 0) & (caret.y < startline) THEN
              DEC(startline,startline - caret.y);
              vscroll.position := startline;
            END;
            IF (caret.y - startline > visiblerows) THEN
              INC(startline,caret.y - startline - visiblerows);
              vscroll.position := startline;
            END;
          END;
        END;

        W.InvalidateRect(handle,NIL,W.False);
      END;
    END MouseMove;

    PROCEDURE MouseUp*(shift: Set; button,x,y: Longint);
    BEGIN
      MouseUp^(shift,button,x,y);
      IF mousepress THEN
        select.end.x := caret.x;
        select.end.y := caret.y;
      END;
      mousepress := FALSE;
      W.InvalidateRect(handle,NIL,W.False);
    END MouseUp;

    PROCEDURE MouseWheel*(shift: Set; delta: Longint; x,y: Longint);
    BEGIN
      MouseWheel^(shift,delta,x,y);
      IF multiline THEN
        startline := startline - 4*delta;
        IF startline > lines.count - 1 - visiblerows THEN startline := lines.count - 1 - visiblerows END;
        IF startline < 0 THEN startline := 0 END;
        vscroll.position := startline;
      END;
      W.InvalidateRect(handle,NIL,W.False);
    END MouseWheel;

    PROCEDURE ShowCaret;
    VAR
      rect : W.Rect;
      coord: W.Point;
    BEGIN
      W.GetClientRect(handle,rect);
      IF multiline THEN
        IF vscroll.visible THEN DEC(rect.right,vscroll.width) END;
      END;
      coord := CaretToCoord(caret.x,caret.y);
      IF U.XYInRect(coord.x,coord.y + lineheight - caretoffset - 1,rect) THEN
        W.SetCaretPos(coord.x,coord.y);
        W.ShowCaret(handle);
        caretvisible := TRUE;
      END;
    END ShowCaret;

    PROCEDURE HideCaret;
    BEGIN
      IF caretvisible THEN
        W.HideCaret(handle);
        caretvisible := FALSE;
      END;
    END HideCaret;

    PROCEDURE CaretToCoord(col,row: Longint): W.Point;
    VAR
      i,w: Longint;
      res: W.Point;
      c  : Char;
      f  : Shortint;
    BEGIN
      i := 0;
      w := 0;
      res.x := -offset;
      res.y := caretoffset;

      IF (row >= 0) & (row <= lines.count-1) & (col >= 0) THEN
        f := FontStyleToInt(canvas.font);
        WHILE i < col DO
          c := lines.CharByPos(i,row);
          w := w + charwidths[f,ORD(c)];
          INC(i);
        END;
        res.x := res.x + w - 1;
        res.y := res.y + (row - startline) * lineheight;
      END;

      IF col <= 0 THEN res.x := -offset; END;
      IF row < 0 THEN res.y := caretoffset END;
      IF row > lines.count-1 THEN res.y := lines.count-1+caretoffset END;

      RETURN res;
    END CaretToCoord;

    PROCEDURE CoordToCaret(x,y: Longint);
    VAR
      w      : Longint;
      col,row: Longint;
      c      : Char;
      f      : Shortint;
    BEGIN
      IF lines.count > 0 THEN
        col := 0;
        x := x + offset;
        row := y DIV lineheight + startline;
        IF row < 0 THEN row := 0 END;
        IF row > lines.count - 1 THEN row := lines.count - 1 END;
        c := lines.CharByPos(col,row);
        w := 0;
        f := FontStyleToInt(canvas.font);
        WHILE (w < x - charwidths[f,ORD(c)]) DO
          c := lines.CharByPos(col,row);
          w := w + charwidths[f,ORD(c)];
          INC(col);
        END;
        caret.x := col;
        c:=lines.CharByPos(col,row);
        IF x - w >= charwidths[f,ORD(c)] DIV 2 THEN INC(caret.x) END;
        caret.y := row;
      END;
    END CoordToCaret;

    PROCEDURE Resize*;
    VAR
      w,h: Longint;
    BEGIN
      Resize^;
      w := 0;
      h := 0;

      IF multiline THEN
        IF vscroll.visible THEN w := vscroll.width END;
      END;

      canvas.SetSize(width-w,height-h);
      visiblerows      := canvas.height DIV lineheight - 1;

      IF multiline THEN
        vscroll.pagesize := visiblerows;
        IF (visiblerows > endline-startline) & (endline = lines.count - 1) & (startline > 0) THEN
          DEC(startline,visiblerows - (endline-startline));
          vscroll.position := startline;
        END;
      END;
    END Resize;

    PROCEDURE WMEraseBkgnd*(VAR message: Cl.Message);
    BEGIN
      message.result := 1;
    END WMEraseBkgnd;

    PROCEDURE WMPaint*(VAR message: Cl.Message);
    VAR
      dc   : W.HDC;
    BEGIN
      HideCaret;
      dc := W.GetDC(handle);
      Paint;
      canvas.Draw(dc,0,0);
      W.ReleaseDC(handle,dc);
      ShowCaret;
    END WMPaint;

    PROCEDURE ClearSelection*;
    BEGIN
      select.start.x:= -1;
      select.start.y:= -1;
      select.end.x  := -1;
      select.end.y  := -1;
      select.count := 0;
    END ClearSelection;

    PROCEDURE KeyPress*(key: Char);
    VAR
      i: Longint;
      start : W.Point;
      end   : W.Point;
    BEGIN
      KeyPress^(key);

      IF select.count > 0 THEN

        start := select.start;
        end   := select.end;

        IF start.y > end.y THEN
          U.Swap(start.y,end.y);
          U.Swap(start.x,end.x);
        ELSIF start.y = end.y THEN
          IF start.x > end.x THEN U.Swap(start.x,end.x) END;
        END;

        lines.DeleteRange(start,end);
        select.count := 0;
        caret.x := start.x;
        caret.y := start.y;

      END;

      CASE key OF
      |  20X..0FFX : lines.InsertChar(caret.x,caret.y,key); INC(caret.x);
      ELSE
      END;

      ClearSelection;

      IF multiline THEN
        vscroll.max := lines.count - 1;
        vscroll.pagesize := visiblerows;
        vscroll.position := startline;
      END;
      W.InvalidateRect(handle,NIL,W.False);
    END KeyPress;

    PROCEDURE KeyDown*(key: Shortint; shift: Set);
    VAR
      start : W.Point;
      end   : W.Point;
      l     : Longint;
      s:ARRAY 1024 OF Char;
    BEGIN
      KeyDown^(key,shift);

      IF select.count > 0 THEN
        start := select.start;
        end   := select.end;
        IF start.y > end.y THEN
          U.Swap(start.y,end.y);
          U.Swap(start.x,end.x);
        ELSIF start.y = end.y THEN
          IF start.x > end.x THEN U.Swap(start.x,end.x) END;
        END;
      END;

      CASE key OF
      |  W.VK_LEFT  : DEC(caret.x); select.count := 0;
      |  W.VK_RIGHT : INC(caret.x); select.count := 0;
      |  W.VK_UP    : DEC(caret.y); select.count := 0;
      |  W.VK_DOWN  : INC(caret.y); select.count := 0;
      |  W.VK_DELETE:
                      IF select.count > 0 THEN
                        lines.DeleteRange(start,end);
                        select.count := 0;
                        caret.x := start.x;
                        caret.y := start.y;
                      ELSE
                        l := lines.DeleteChar(caret.x,caret.y);
                        IF (l = 0) & (caret.y # lines.count -1) THEN
                          lines.JoinLine(caret.x,caret.y+1)
                        END;
                      END;
      |  W.VK_RETURN:
                      IF select.count > 0 THEN
                        lines.DeleteRange(start,end);
                        select.count := 0;
                        caret.x := start.x;
                        caret.y := start.y;
                      END;
                      IF multiline THEN
                        lines.SplitLine(caret.x,caret.y);
                        caret.x := 0;
                        INC(caret.y);
                      END;
      |  W.VK_BACK  :
                      IF select.count > 0 THEN
                        lines.DeleteRange(start,end);
                        select.count := 0;
                        caret.x := start.x;
                        caret.y := start.y;
                      ELSE
                        l := lines.DeleteChar(caret.x-1,caret.y);
                        DEC(caret.x);
                        IF caret.x < 0 THEN
                          caret.x := lines.Length(caret.y-1);
                          lines.JoinLine(-1,caret.y);
                          DEC(caret.y);
                        END;
                      END;

      | W.VK_F1:

            FOR l := 0 TO 250 DO
              s := "   WEDA soft!!!    WEDA soft!!!   ";
              U.Append(U.IntToStr(l),s);
              lines.Add(s);
            END;

      ELSE
      END;

      IF caret.x < 0 THEN caret.x := 0 END;
      IF (startline > 0) & (caret.y < startline) THEN DEC(startline,startline - caret.y) END;
      IF caret.y < 0 THEN caret.y := 0 END;
      IF caret.y > lines.count - 1 THEN caret.y := lines.count - 1 END;
      IF (caret.y - startline > visiblerows) THEN INC(startline,caret.y - startline - visiblerows) END;
      IF multiline THEN
        vscroll.position := startline;
        vscroll.max := lines.count - 1;
        (*vscroll.visible := visiblerows < lines.count - 1;*)
      END;
      W.InvalidateRect(handle,NIL,W.False);
    END KeyDown;

    PROCEDURE DrawSelect(line: Longint);
    VAR
      rect  : W.Rect;
      start : W.Point;
      end   : W.Point;
      old,c : Longint;
      coord1: W.Point;
      coord2: W.Point;
    BEGIN

      start := select.start;
      end   := select.end;

      IF start.y > end.y THEN
        U.Swap(start.y,end.y);
        U.Swap(start.x,end.x);
      ELSIF start.y = end.y THEN
        IF start.x > end.x THEN U.Swap(start.x,end.x) END;
      END;

      IF (select.count > 0 ) & (line - startline >= start.y - startline) & (line - startline <= end.y - startline) THEN

        IF line = caret.y THEN
          c := G.clWebLightYellow;
        ELSE
          c := color;
        END;

        coord1 := CaretToCoord(start.x,start.y);
        coord2 := CaretToCoord(end.x,end.y);

        IF (line - startline = start.y - startline) & (line - startline = end.y - startline) THEN

          canvas.FillRect(0,(line - startline) * lineheight,coord2.x+1,(line - startline) * lineheight + lineheight,c);
          canvas.TextOut (-offset,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.SubString(0,start.x,line));

          IF coord1.x = 0 THEN DEC(coord1.x) END;
          canvas.FillRect(coord1.x+1,(line - startline) * lineheight,coord2.x+1,(line - startline) * lineheight + lineheight,select.color);
          old := canvas.font.color;
          canvas.font.color := select.font.color;
          canvas.TextOut (-offset+coord1.x+1,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.SubString(start.x,end.x,line));
          canvas.font.color := old;

          IF coord2.x = 0 THEN DEC(coord2.x) END;
          canvas.FillRect(coord2.x+1,(line - startline) * lineheight,width,(line - startline) * lineheight + lineheight,c);
          canvas.TextOut (-offset+coord2.x+1,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.SubString(end.x,-1,line));

        ELSIF line - startline = start.y - startline THEN

          canvas.FillRect(0,(line - startline) * lineheight,coord1.x,(line - startline) * lineheight + lineheight,c);
          canvas.TextOut (-offset,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.SubString(0,start.x,line));

          IF coord1.x = 0 THEN DEC(coord1.x) END;
          canvas.FillRect(coord1.x+1,(line - startline) * lineheight,width,(line - startline) * lineheight + lineheight,select.color);
          old := canvas.font.color;
          canvas.font.color := select.font.color;
          canvas.TextOut (coord1.x+1,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.SubString(start.x,-1,line));
          canvas.font.color := old;

        ELSIF line - startline < end.y - startline THEN

          canvas.FillRect(0,(line - startline) * lineheight,canvas.width,(line - startline) * lineheight + lineheight,select.color);
          old := canvas.font.color;
          canvas.font.color := select.font.color;
          canvas.TextOut (-offset,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.Strings(line));
          canvas.font.color := old;

        ELSIF line - startline = end.y - startline THEN
          IF coord2.x # 0 THEN
            canvas.FillRect(0,(line - startline) * lineheight,coord2.x+1,(line - startline) * lineheight + lineheight,select.color);
            old := canvas.font.color;
            canvas.font.color := select.font.color;
            canvas.TextOut (-offset,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.SubString(0,end.x-1,line));
            canvas.font.color := old;
          END;
          IF coord2.x = 0 THEN DEC(coord2.x) END;
          canvas.FillRect(coord2.x+1,(line - startline) * lineheight,width,(line - startline) * lineheight + lineheight,c);
          canvas.TextOut (coord2.x+1,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.SubString(end.x,-1,line));

        END;

      ELSE
        IF line = caret.y THEN
          canvas.FillRect(0,(line - startline) * lineheight,canvas.width,canvas.height-1,G.clWebLightYellow);
        ELSE
          canvas.FillRect(0,(line - startline) * lineheight,canvas.width,(line - startline) * lineheight + lineheight,color);
        END;
        canvas.TextOut (-offset,(line - startline) * lineheight + (lineheight - charheight) DIV 2 - 2,lines.Strings(line));
      END;

    END DrawSelect;

    PROCEDURE Paint*;
    VAR
      i: Longint;
    BEGIN
      endline := canvas.height DIV lineheight + startline - 1;
      IF endline > lines.count - 1 THEN endline := lines.count - 1 END;

      IF ~ multiline THEN
        startline := 0;
        endline   := 0;
      END;

      FOR i := startline TO endline DO
        DrawSelect(i);
        IF i = caret.y THEN currentline := i END;
      END;
      canvas.FillRect(0,(i - startline) * lineheight,canvas.width,canvas.height,color);
    END Paint;

    PROCEDURE WindowProc*(VAR message: Cl.Message);
    BEGIN
      CASE message.msg OF
      |  W.WM_GETDLGCODE     : message.result := SYSTEM.VAL(Longint,W.DLGC_WANTARROWS + W.DLGC_WANTALLKEYS + W.DLGC_WANTTAB);
      ELSE
        WindowProc^(message);
      END;
    END WindowProc;

  END Customedit;

  Edit* = OBJECT(Customedit) END Edit;

  Memo* = OBJECT(Customedit)

    PROCEDURE & Create*(Parent: C.Control);
    BEGIN
      Create^(Parent);
      multiline      := TRUE;
    END Create;

    PROCEDURE LoadFromFile*(name: String);
    BEGIN
      W.SetFocus(handle);
      lines.LoadFromFile(name);
      ClearSelection;
      caret.x := 0;
      caret.y := 0;
      Calculate;
      W.InvalidateRect(handle,NIL,W.False);
    END LoadFromFile;

  END Memo;

(*################################################*)
(*################################################*)

END Editors.

