MODULE Menus;

IMPORT
  SYSTEM, Trace, W := Windows, Cl := Classes, G := Graphics,
  U := Utilites, A := Assistants;

TYPE
  Byte         = W.Byte;
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

  ItemEvent*     = PROCEDURE {DELEGATE} (Sender: Menuitem);

  Menuitem* = OBJECT
  VAR
    tag*       : Longint;
    caption*   : Name;
    enabled*   : Boolean;
    imageindex*: Longint;
    popup*     : Popupmenu;
    control*   : Popupmenu;
    rect*      : W.Rect;
    mode*      : Shortint;
  END Menuitem;

  Forbutton* = RECORD (G.Pack)
    frame* : G.Pack;
    font*  : G.StdPack;
  END;

  Forshadow* = RECORD
    color* : Longint;
    light* : Longint;
    night* : Longint;
  END;

  Colorfor* = RECORD
    button*: Forbutton;
    shadow*: Forshadow;
    popup* : Forshadow;
  END;

  Popupmenu* = OBJECT (Cl.Popupmenu)
  VAR
    items*      : Cl.List;
    imagelist*  : G.Imagelist;
    gutter*     : A.Gutter;
    itemheight* : Longint;
    current*    : Menuitem;
    colorfor*   : Colorfor;

    linewidth*  : Longint;
    linecolor*  : Longint;

    OnItemClick*: ItemEvent;
    OnPopup*    : Cl.NotifyEvent;

    activate    : Boolean;
    level       : Longint;

    PROCEDURE & InitPopup*(AOwner: Cl.Wincontrol);
    BEGIN
      InitPopup^(AOwner);

      NEW(gutter);
      property.AddObject("gutter",gutter);
      property.Sort;

      itemheight     := 26;
      linewidth      := 0;
      activate       := FALSE;
      NEW(items,0);

      colorfor.popup.color          := G.clBtnFace;
      colorfor.button.color         := G.clWhite;
      colorfor.button.actived       := G.clHiGray50;(*0EFD3C6H;*)
      colorfor.button.frame.color   := G.clHiGray35;
      colorfor.button.frame.actived := G.clHiGray30;(*0C66931H;*)
      colorfor.button.font.enabled  := G.clGray60;
      colorfor.button.font.disabled := G.clGray90;
      colorfor.button.font.actived  := G.clGray35;

      colorfor.shadow.color         := G.clHiGray35;
      colorfor.shadow.light         := G.clWhite;
      colorfor.shadow.night         := G.clGray40;

      level          := 0;
      IF control IS Popupmenu THEN
        level := control(Popupmenu).level + 1;
      END;
    END InitPopup;

    PROCEDURE CreateParams*;
    BEGIN
      style    := W.WS_POPUP;
      exstyle  := W.WS_EX_TOOLWINDOW + W.WS_EX_COMPOSITED;
      wndstyle := W.CS_DROPSHADOW - W.CS_VREDRAW - W.CS_HREDRAW - W.CS_DBLCLKS;
      color    := colorfor.popup.color;
    END CreateParams;

    PROCEDURE Destroy*;
    BEGIN
      items.Clear;
      items := NIL;
      canvas.Free;
      gutter.Destroy;
    END Destroy;

    PROCEDURE Hidemain*;
    BEGIN
      IF (control IS Mainmenu) THEN
        IF ~ control(Mainmenu).active THEN
          control(Mainmenu).switch := FALSE;
          control(Mainmenu).current := NIL;
          W.InvalidateRect(control(Mainmenu).handle,NIL,W.False);
        END;
      END;
    END Hidemain;

    PROCEDURE AddItem*(caption: String): Menuitem;
    VAR
      item: Menuitem;
    BEGIN
      NEW(item);
      item.popup := NIL;
      item.imageindex := -1;
      item.mode    := 0;
      item.tag     := 0;
      item.control := SELF;
      item.enabled := TRUE;
      COPY(caption,item.caption);
      items.Add(item);
      RETURN item;
    END AddItem;

    PROCEDURE AddSubItem*(parent: Menuitem; caption: String): Menuitem;
    VAR
      item: Menuitem;
    BEGIN
      IF parent.popup = NIL THEN
        NEW(parent.popup,parent.control);
        parent.popup.OnItemClick := OnItemClick;
        IF imagelist # NIL THEN parent.popup.imagelist := imagelist END;
      END;
      item := parent.popup.AddItem(caption);
      RETURN item;
    END AddSubItem;

    PROCEDURE AddSeparator*(): Menuitem;
    VAR
      item: Menuitem;
    BEGIN
      item := AddItem("-");
      item.mode := 1;
      RETURN item;
    END AddSeparator;

    PROCEDURE Popup*(x,y: Longint);
    VAR
      p: W.Point;
    BEGIN
      current := NIL;
      IF (x # -1) & (y # -1) THEN
        p.x := x;
        p.y := y;
      ELSE
        W.GetCursorPos(p);
      END;

      IF p.x + width > Cl.screen.width THEN
        p.x := Cl.screen.width - width;
      END;

      IF p.y + height > Cl.screen.height THEN
        p.y := Cl.screen.height - height;
      END;

      left := p.x;
      top  := p.y;

      W.SetWindowPos(
        handle, W.HWND_TOPMOST,
        p.x+1,p.y+1,width,height,
        W.SWP_SHOWWINDOW + W.SWP_NOACTIVATE
      );

      IF ~ activate THEN
        W.SetWindowLong(handle,W.GWL_EXSTYLE,W.WS_EX_NOACTIVATE);
        activate := TRUE;
      END;
      visible    := TRUE;
      W.GetWindowRect(handle,Cl.popuprect);

      IF OnPopup # NIL THEN OnPopup(SELF) END;
      W.SetFocus(Cl.activeform);
    END Popup;

    PROCEDURE GetItemAtMousePos*(x,y: Longint): Menuitem;
    VAR
      res : Menuitem;
      i   : Longint;
    BEGIN
      res := NIL;
      i   := 0;
      WHILE (i < items.count) & (res = NIL) DO
        IF (y >= items.list[i](Menuitem).rect.top) & (y <= items.list[i](Menuitem).rect.bottom) THEN
          IF (items.list[i](Menuitem).mode # 1) THEN
            res := items.list[i](Menuitem);
          ELSIF (items.list[i](Menuitem).mode = 1) & (current # NIL) THEN
            res := current;
          END;
        END;
        INC(i);
      END;
      RETURN res;
    END GetItemAtMousePos;

    PROCEDURE MouseLeave*;
    BEGIN
      MouseLeave^;
      IF (current # NIL) & (current.popup = NIL) THEN
        current := NIL;
        HiddenSubPopup(SELF);
        W.InvalidateRect(handle,NIL,W.False);
      END;
    END MouseLeave;

    PROCEDURE HiddenSubPopup*(VAR popup: Popupmenu);
    VAR
      i: Longint;
    BEGIN
      FOR i := 0 TO popup.items.count-1 DO
        IF (popup.items.list[i](Menuitem).popup # NIL) & popup.items.list[i](Menuitem).popup.visible THEN
          HiddenSubPopup(popup.items.list[i](Menuitem).popup);
          IF current # popup.items.list[i](Menuitem) THEN
            popup.items.list[i](Menuitem).popup.Hide;
          END;
        END;
      END;
    END HiddenSubPopup;

    PROCEDURE MouseMove*(shift: Set; x,y: Longint);
    VAR
      res: Menuitem;
      p  : W.Point;
    BEGIN
      MouseMove^(shift,x,y);
      W.GetWindowRect(handle,Cl.popuprect);
      current := GetItemAtMousePos(x,y);
      HiddenSubPopup(SELF);
      IF current # NIL THEN
        IF (current.popup # NIL) & ~ current.popup.visible THEN
          p.x := current.rect.right;
          p.y := current.rect.top;
          W.ClientToScreen(handle,p);

          current.popup.Popup(p.x-1,p.y+1);
        END;
        W.InvalidateRect(handle,NIL,W.False);
      END;
    END MouseMove;

    PROCEDURE MouseUp*(shift: Set; button,x,y: Longint);
    VAR
      i: Menuitem;
    BEGIN
      MouseUp^(shift,button,x,y);
      IF (current # NIL) & current.enabled THEN
        i := current;
        PopupHide(NIL);
        IF control IS Mainmenu THEN
          control(Mainmenu).switch := FALSE;
          control(Mainmenu).current := NIL;
          W.InvalidateRect(control(Mainmenu).handle,NIL,W.False);
        END;
        current := NIL;
        visible := FALSE;
        current := i;
        Click;
      END;
      W.GetWindowRect(control.handle,Cl.popuprect);
    END MouseUp;

    PROCEDURE Resize*;
    BEGIN
      IF handle # 0 THEN
        canvas.SetSize(width,height);
      END;
    END Resize;

    PROCEDURE Click*;
    BEGIN
      IF (current # NIL) & (OnItemClick # NIL) THEN OnItemClick(current) END;
    END Click;

    PROCEDURE DrawItems(dc: W.HDC);
    VAR
      i,x,y,h,c: Longint;
    BEGIN
      canvas.FillRect(0,0,gutter.width,height,G.clWhite);
      canvas.Rectangle(0,0,width-1,height-1,colorfor.button.frame.color);
      IF linewidth > 0 THEN
        canvas.Line(1,0,linewidth+2,0,linecolor);
      END;
      canvas.Line(gutter.width,0,gutter.width,height-1,gutter.linecolor);
      canvas.Line(gutter.width+1,1,gutter.width+1,height-2,gutter.lineshadow);
      canvas.font.color := colorfor.button.font.enabled;

      IF (current # NIL) & current.enabled THEN
        canvas.Roundrect(current.rect.left - gutter.width + 2,current.rect.top + 2,current.rect.right - 2,current.rect.bottom - 2,2,colorfor.button.frame.actived,colorfor.button.actived);
        canvas.Line(current.rect.left-1,current.rect.top+3,current.rect.left-1,current.rect.bottom - 3,colorfor.shadow.color);
        canvas.Line(current.rect.left,current.rect.top+3,current.rect.left,current.rect.bottom - 3,colorfor.shadow.light);
      END;

      FOR i := 0 TO items.count -1 DO
        IF (imagelist # NIL) & (items.list[i](Menuitem).mode # 1) THEN
          IF items.list[i](Menuitem).imageindex # -1 THEN
            x := (gutter.width - imagelist.width) DIV 2;
            y := items.list[i](Menuitem).rect.top + (itemheight - imagelist.height) DIV 2;
            IF items.list[i](Menuitem).enabled THEN
              c := -1;
            ELSE
              c := colorfor.shadow.night;
            END;
            canvas.DrawImage(x,y,imagelist.GetImage(items.list[i](Menuitem).imageindex),c);
          END;
        END;
        IF items.list[i](Menuitem).mode = 0 THEN
          IF (current # NIL) & (current = items.list[i](Menuitem)) THEN
            canvas.font.color := colorfor.button.font.actived;
            IF ~ current.enabled THEN canvas.font.color := colorfor.button.font.disabled END;
          ELSE
            IF items.list[i](Menuitem).enabled THEN
              c := colorfor.button.font.enabled;
            ELSE
              c := colorfor.button.font.disabled;
            END;
            canvas.font.color := c;
          END;
          canvas.TextOut(items.list[i](Menuitem).rect.left+5,items.list[i](Menuitem).rect.top + (itemheight - canvas.font.height - 3) DIV 2 - 1,items.list[i](Menuitem).caption);
        ELSIF items.list[i](Menuitem).mode = 1 THEN
          canvas.Line(items.list[i](Menuitem).rect.left + 2,items.list[i](Menuitem).rect.top + 2,items.list[i](Menuitem).rect.right - 2,items.list[i](Menuitem).rect.top + 2,colorfor.shadow.color);
          canvas.Line(items.list[i](Menuitem).rect.left + 2,items.list[i](Menuitem).rect.top + 3,items.list[i](Menuitem).rect.right - 2,items.list[i](Menuitem).rect.top + 3,colorfor.shadow.light);
        END;
        IF items.list[i](Menuitem).popup # NIL THEN
          x := items.list[i](Menuitem).rect.right - 10;
          y := items.list[i](Menuitem).rect.top + itemheight DIV 2;
          c := canvas.font.color;
          canvas.Line(x,y-4,x,y+4,    c);
          canvas.Line(x+1,y-3,x+1,y+3,c);
          canvas.Line(x+2,y-2,x+2,y+2,c);
          canvas.Line(x+3,y-1,x+3,y+1,c);
        END;
      END;
    END DrawItems;

    PROCEDURE WMPaint*(VAR message: Cl.Message);
    VAR
      dc     : W.HDC;
      i,w,h,s: Longint;
      size   : W.Size;
    BEGIN
      dc := W.GetDC(handle);

      W.SelectObject(dc,canvas.font.handle);
      w := 0;
      h := 0;
      FOR i := 0 TO items.count -1 DO
        s := itemheight;
        IF items.list[i](Menuitem).mode = 1 THEN s := 4 END;
        items.list[i](Menuitem).rect.left   := gutter.width;
        items.list[i](Menuitem).rect.top    := h;
        W.GetTextExtentPoint32(dc,SYSTEM.ADR(items.list[i](Menuitem).caption),U.Length(items.list[i](Menuitem).caption),size);
        items.list[i](Menuitem).rect.right  := items.list[i](Menuitem).rect.left + 5 + size.cx + 20;
        items.list[i](Menuitem).rect.bottom := items.list[i](Menuitem).rect.top + s;
        IF w < gutter.width + 5 + size.cx + 20 THEN w := gutter.width + 5 + size.cx + 20 END;
        h := items.list[i](Menuitem).rect.bottom;
      END;
      INC(h);

      IF w < gutter.width*2 THEN w := gutter.width*2 END;
      FOR i := 0 TO items.count -1 DO
        items.list[i](Menuitem).rect.right  := items.list[i](Menuitem).rect.left + w - gutter.width;
      END;

      IF (w # width) OR (h # height) THEN
        width  := w;
        height := h;
        W.SetWindowPos(
          handle, W.HWND_TOPMOST,
          0, 0, width, height,
          W.SWP_NOMOVE
        );
      END;

      canvas.FillRect(0,0,width,height,color);
      DrawItems(dc);
      canvas.Draw(dc,0,0);
      W.ReleaseDC(handle,dc);
    END WMPaint;

    PROCEDURE WindowProc*(VAR message: Cl.Message);
    BEGIN
      CASE message.msg OF
      |  W.WM_LBUTTONDOWN : IF W.GetFocus() # Cl.activeform THEN W.SetFocus(Cl.activeform) END;
      |  W.WM_SETFOCUS    : IF W.GetFocus() # Cl.activeform THEN W.SetFocus(Cl.activeform) END;
      ELSE
        WindowProc^(message);
      END;
    END WindowProc;

  END Popupmenu;

  Mainmenu* = OBJECT (Cl.Mainmenu)
  VAR
    items        : Cl.List;
    current      : Menuitem;
    switch       : Boolean;
    active       : Boolean;
    menuheight   : Longint;

    colorfor*    : Colorfor;
    imagelist*   : G.Imagelist;
    OnMenuClick* : ItemEvent;
    OnPopup*     : Cl.NotifyEvent;

    PROCEDURE & Create*(Parent: Cl.Wincontrol);
    BEGIN
      Create^(Parent);
      classname  := "Mainmenu";
      left       := 0;
      top        := 0;
      width      := 0;
      height     := 0;
      menuheight := 27;
      align      := Cl.alTop;
      color      := G.clWhite;
      switch     := FALSE;
      active     := FALSE;
      current    := NIL;

      colorfor.popup.color          := G.clBtnFace;
      colorfor.button.color         := G.clWhite;
      colorfor.button.actived       := G.clHiGray50;(*0EFD3C6H;*)
      colorfor.button.frame.color   := G.clHiGray35;
      colorfor.button.frame.actived := G.clHiGray30;(*0C66931H;*)
      colorfor.button.font.enabled  := G.clGray60;
      colorfor.button.font.disabled := G.clGray90;
      colorfor.button.font.actived  := G.clGray35;
      colorfor.shadow.color         := G.clHiGray35;
      colorfor.shadow.light         := color;

      NEW(items,0);

    END Create;

    PROCEDURE Destroy*;
    BEGIN

      Destroy^;
    END Destroy;

    PROCEDURE AdjustRect*;
    VAR
      index: Longint;
    BEGIN
      AdjustRect^;
      IF owner # NIL THEN
        index := owner.controls.IndexOf(SELF);
        IF index # 0 THEN
          owner.controls.SwapByIndex(0,index);
        END;
      END;
    END AdjustRect;

    PROCEDURE GetItemAtMousePos*(x,y: Longint): Menuitem;
    VAR
      res : Menuitem;
      i   : Longint;
    BEGIN
      res := NIL;
      i   := 0;
      WHILE (i < items.count) & (res = NIL) DO
        IF (x >= items.list[i](Menuitem).rect.left) & (x <= items.list[i](Menuitem).rect.right) THEN
          res := items.list[i](Menuitem);
        END;
        INC(i);
      END;
      IF ~ mousedown & switch & (current # NIL) & (res = NIL) THEN res := current END;
      RETURN res;
    END GetItemAtMousePos;

    PROCEDURE HiddenPopup*(VAR popup: Popupmenu);
    VAR
      i: Longint;
    BEGIN
      IF Cl.screen # NIL THEN
        FOR i := 0 TO Cl.screen.count -1 DO
          IF (Cl.screen.list[i] IS Popupmenu) & (popup # Cl.screen.list[i](Popupmenu)) & Cl.screen.list[i](Popupmenu).visible THEN
            Cl.screen.list[i](Popupmenu).visible := FALSE;
            W.ShowWindow(Cl.screen.list[i](Popupmenu).handle,W.SW_HIDE);
          END;
        END;
      END;
    END HiddenPopup;

    PROCEDURE MouseLeave*;
    BEGIN
      MouseLeave^;
      active     := FALSE;
      IF ~ switch THEN
        current := NIL;
        W.InvalidateRect(handle,NIL,W.False);
      ELSIF (current # NIL) & (current.popup = NIL) THEN
        current := NIL;
        switch  := FALSE;
        W.InvalidateRect(handle,NIL,W.False);
      END;
    END MouseLeave;

    PROCEDURE MouseMove*(shift: Set; x,y: Longint);
    VAR
      p  : W.Point;
    BEGIN
      MouseMove^(shift,x,y);
      active     := TRUE;
      current := GetItemAtMousePos(x,y);
      IF (current # NIL) & (current.popup # NIL) THEN
        current.popup.current := NIL;
        W.InvalidateRect(current.popup.handle,NIL,W.False);
      END;
      IF switch & (current # NIL) THEN
        HiddenPopup(current.popup);
        IF (current.popup # NIL) & ~ current.popup.visible THEN
          p.x := current.rect.left;
          p.y := current.rect.bottom;
          W.ClientToScreen(handle,p);
          current.popup.Popup(p.x - 2,p.y - 1);
        END;
      END;
      IF switch & (current = NIL) THEN
        PopupHide(NIL);
      END;
      W.InvalidateRect(handle,NIL,W.False);
    END MouseMove;

    PROCEDURE MouseDown*(shift: Set; button,x,y: Longint);
    VAR
      p  : W.Point;
    BEGIN
      MouseDown^(shift,button,x,y);
      switch := ~ switch;
      current := GetItemAtMousePos(x,y);
      IF current = NIL THEN
        switch := FALSE;
      END;
      IF switch & (current # NIL) THEN
        HiddenPopup(current.popup);
        IF (current.popup # NIL) & ~ current.popup.visible THEN
          p.x := current.rect.left;
          p.y := current.rect.bottom;
          W.ClientToScreen(handle,p);
          current.popup.Popup(p.x - 2,p.y - 1);
          IF OnPopup # NIL THEN OnPopup(current.popup) END;
        END;
      END;
      (*Cl.activeform := handle;*)
      W.InvalidateRect(handle,NIL,W.False);
    END MouseDown;

    PROCEDURE MouseUp*(shift: Set; button,x,y: Longint);
    BEGIN
      MouseUp^(shift,button,x,y);
      IF current # NIL THEN MenuClick(current) END;
    END MouseUp;

    PROCEDURE MenuClick(Sender: Menuitem);
    VAR
      i: Menuitem;
    BEGIN
      i := current;
      current := NIL;
      W.InvalidateRect(handle,NIL,W.False);
      IF i # Sender THEN
        switch := FALSE;
        IF OnMenuClick # NIL THEN OnMenuClick(Sender) END;
      END;
    END MenuClick;

    PROCEDURE AddMenu*(caption: String): Menuitem;
    VAR
      item: Menuitem;
    BEGIN
      IF height = 0 THEN height := menuheight END;
      NEW(item);
      item.popup := NIL;
      item.imageindex := -1;
      item.mode       := 2;
      item.tag        := 0;
      item.enabled    := TRUE;
      COPY(caption,item.caption);
      items.Add(item);
      RETURN item;
    END AddMenu;

    PROCEDURE AddSubMenu*(parent: Menuitem; caption: String): Menuitem;
    VAR
      item: Menuitem;
    BEGIN
      IF parent.popup = NIL THEN
        NEW(parent.popup,SELF);
        parent.popup.colorfor := colorfor;
        parent.popup.OnItemClick := MenuClick;
        IF imagelist # NIL THEN parent.popup.imagelist := imagelist END;
      END;
      item := parent.popup.AddItem(caption);
      RETURN item;
    END AddSubMenu;

    PROCEDURE AddSeparator*(parent: Menuitem): Menuitem;
    VAR
      item: Menuitem;
    BEGIN
      item := NIL;
      IF (parent # NIL) & (parent.popup # NIL) THEN
        item := parent.popup.AddSeparator();
        item.mode := 1;
      END;
      RETURN item;
    END AddSeparator;

    PROCEDURE Paint*;
    VAR
      i,c1,c2: Longint;
    BEGIN
      IF (current # NIL) & current.enabled THEN
        IF switch THEN
          c1 := colorfor.button.frame.color;
          c2 := colorfor.button.color;
        ELSE
          c1 := colorfor.button.frame.actived;
          c2 := colorfor.button.actived;
        END;
        canvas.FrameRect(current.rect.left,current.rect.top,current.rect.right,current.rect.bottom,c1,c2);
      END;
      FOR i := 0 TO items.count -1 DO

        IF (current # NIL) & current.enabled & (current = items.list[i](Menuitem)) THEN
          canvas.font.color := colorfor.button.font.actived;
        ELSIF items.list[i](Menuitem).enabled THEN
          canvas.font.color := colorfor.button.font.enabled;
        ELSE
          canvas.font.color := colorfor.button.font.disabled;
        END;
        canvas.TextOut(items.list[i](Menuitem).rect.left+5,items.list[i](Menuitem).rect.top + (menuheight - canvas.font.height - 5) DIV 2 - 3,items.list[i](Menuitem).caption);
      END;

      canvas.GradientLine(0,height-1,width DIV 2,height-1,colorfor.shadow.light,colorfor.shadow.color);
      canvas.GradientLine(width DIV 2,height-1,width,height-1,colorfor.shadow.color,colorfor.shadow.light);

    END Paint;

    PROCEDURE WMPaint*(VAR message: Cl.Message);
    VAR
      dc  : W.HDC;
      i,w : Longint;
      size: W.Size;
    BEGIN
      dc := W.GetDC(handle);
      W.SelectObject(dc,canvas.font.handle);
      canvas.FillRect(0,0,width,height,color);
      w := 2;
      FOR i := 0 TO items.count -1 DO
        items.list[i](Menuitem).rect.left   := w;
        items.list[i](Menuitem).rect.top    := 2;
        W.GetTextExtentPoint32(dc,SYSTEM.ADR(items.list[i](Menuitem).caption),U.Length(items.list[i](Menuitem).caption),size);
        items.list[i](Menuitem).rect.right  := items.list[i](Menuitem).rect.left + 5 + size.cx + 5 - 1;
        items.list[i](Menuitem).rect.bottom := items.list[i](Menuitem).rect.top + menuheight - 5 - 1;
        w := items.list[i](Menuitem).rect.right + 1;
        IF (items.list[i](Menuitem).mode = 2) & (items.list[i](Menuitem).popup # NIL) THEN
          items.list[i](Menuitem).popup.linewidth := items.list[i](Menuitem).rect.right - items.list[i](Menuitem).rect.left;
          items.list[i](Menuitem).popup.linecolor := color;
        END;
      END;
      Paint;
      canvas.Draw(dc,0,0);
      W.ReleaseDC(handle,dc);
    END WMPaint;

  END Mainmenu;

END Menus.
