MODULE Syntax;

IMPORT
  SYSTEM, Trace, W := Windows, F := Files, U := Utilites, Cl := Classes,
  C := Controls, G := Graphics, S := Stdctrls;

TYPE
  Byte         = W.Byte;
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

  Gutterline* = RECORD (G.SmallPack)
    width* : Longint;
  END;

  Gutter* = RECORD
    color* : Longint;
    width* : Longint;
    line*  : Gutterline;
  END;

  Attributes* = OBJECT
  VAR
    style*      : Set;
    background* : Longint;
    foreground* : Longint;
  END Attributes;

  Tokenlist* = OBJECT
  VAR
    left   : Tokenlist;
    right  : Tokenlist;
    name   : Name;
    count* : Longint;

    PROCEDURE & Init*;
    BEGIN
      left    := NIL;
      right   := NIL;
      name    := "";
      count   := 0;
    END Init;

    PROCEDURE Add*(name: Name);
    VAR
      ob0, ob1: Tokenlist;
      left: Boolean;
    BEGIN

      ob0 := SELF;
      ob1 := ob0.right;
      left := FALSE;

      LOOP
        IF ob1 # NIL THEN
          IF name < ob1.name THEN
	    ob0 := ob1;
	    ob1 := ob0.left;
	    left := TRUE
          ELSIF name > ob1.name THEN
	    ob0 := ob1;
	    ob1 := ob0.right;
	    left := FALSE
          ELSE
            EXIT
          END;
        ELSE (*insert*)
          NEW(ob1);
          IF left THEN
	    ob0.left := ob1
          ELSE
	    ob0.right := ob1
          END;
          ob1.left:=NIL;
          ob1.right:=NIL;
          ob1.name := name;
          INC(count);
          EXIT;
        END;
      END;

    END Add;

    PROCEDURE Find*(name: String): Tokenlist;
    VAR
      obj: Tokenlist;
      res: Name;
    BEGIN
      obj := SELF;
      LOOP
        IF obj # NIL THEN
          IF name < obj.name THEN
	    obj := obj.left;
          ELSIF name > obj.name THEN
	    obj := obj.right;
          ELSE (*ok*)
            EXIT
          END;
        ELSE (*no found*)
          RETURN NIL;
        END;
      END;
      RETURN obj;
    END Find;

    PROCEDURE Free*;

      PROCEDURE FreeObj(VAR obj: Tokenlist);
      BEGIN
        IF obj = NIL THEN RETURN END;
        IF obj.left  # NIL THEN FreeObj(obj.left); END;
        IF obj.right # NIL THEN FreeObj(obj.right); END;

        obj := NIL;
      END FreeObj;

    BEGIN
      FreeObj(left);
      FreeObj(right);
    END Free;

  END Tokenlist;

  Basehighlighter* = OBJECT
  VAR
    memo*     : Memo;
    tokenlist*: Tokenlist;

    PROCEDURE & Create*(Parent: Memo);
    BEGIN
      memo := Parent;
      NEW(tokenlist);
    END Create;

    PROCEDURE Free*;
    BEGIN
      tokenlist.Free;
    END Free;

  END Basehighlighter;

  Memo* = OBJECT(C.Control)
  VAR
    lines*       : Cl.Stringlist;
    canvas*      : G.Bitmap;
    gutter*      : Gutter;
    highlighter* : Basehighlighter;
    crdefault    : W.HCursor;
    crbeam       : W.HCursor;

    PROCEDURE & Create*(Parent: C.Control);
    BEGIN
      Create^(Parent);
      classname      := "Synedit";
      width          := 350;
      height         := 250;
      color          := G.clWhite;

      gutter.width      := 60;
      gutter.color      := G.clHiGray50;
      gutter.line.width := 1;
      gutter.line.color := G.clHiGray30;
      gutter.line.shadow:= G.clWhite;

      NEW(canvas);
      NEW(lines);
      lines.Add("first");

    END Create;

    PROCEDURE Destroy*;
    BEGIN
      canvas.Free;
      canvas := NIL;
      lines.Free;
      lines   := NIL;
      Destroy^;
    END Destroy;

    PROCEDURE CreateParams*;
    BEGIN
      CreateParams^;
      style       := style - W.WS_BORDER;
      wndstyle    := wndstyle + W.CS_VREDRAW + W.CS_HREDRAW;
      crdefault   := W.LoadCursor(0,SYSTEM.VAL(W.LPSTR,W.IDC_IBEAM));
      crbeam      := W.LoadCursor(0,SYSTEM.VAL(W.LPSTR,W.IDC_IBEAM));
      canvas.SetSize(width,height);
    END CreateParams;

    PROCEDURE Resize*;
    BEGIN
      Resize^;
      canvas.SetSize(width,height);
    END Resize;

    PROCEDURE WMPaint*(VAR message: Cl.Message);
    VAR
      dc   : W.HDC;
    BEGIN
      (*HideCaret;*)
      dc := W.GetDC(handle);
      Paint;
      canvas.Draw(dc,0,0);
      W.ReleaseDC(handle,dc);
      (*ShowCaret;*)
    END WMPaint;

    PROCEDURE Paint*;
    BEGIN

      canvas.FillRect(0,0,gutter.width,height,gutter.color);
      canvas.Line(gutter.width,0,gutter.width,height,gutter.line.color);
      canvas.Line(gutter.width+1,0,gutter.width+1,height,gutter.line.shadow);
      canvas.FillRect(gutter.width+2,0,gutter.width+12,height,gutter.color);

      IF highlighter # NIL THEN

      ELSE
        canvas.TextOut (gutter.width+12+2,0,lines.Strings(0));
      END;

    END Paint;

  END Memo;

END Syntax.

