MODULE Utilites;

IMPORT
  SYSTEM, W := Windows;

TYPE
  Byte         = W.Byte;
  Char         = W.Char;
  Boolean      = W.Boolean;
  Shortint     = W.Shortint;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Real         = W.Real;
  Longreal     = W.Longreal;
  Pointer      = W.Pointer;
  Set	       = W.Set;
  String       = W.String;
  Name         = W.Name;

VAR
  ZRand: Longint;

PROCEDURE Swap*(VAR x,y: Longint);
VAR
  t: Longint;
BEGIN
  t := x;
  x := y;
  y := t;
END Swap;

PROCEDURE Rect*(left,top,right,bottom: Longint): W.Rect;
VAR
  rect: W.Rect;
BEGIN
  rect.left   := left;
  rect.top    := top;
  rect.right  := right;
  rect.bottom := bottom;
  RETURN rect;
END Rect;

PROCEDURE IsRectEmpty*(rect: W.Rect): Boolean;
BEGIN
  RETURN (rect.right <= rect.left) OR (rect.bottom <= rect.top);
END IsRectEmpty;

PROCEDURE XYInRect*(x,y: Longint; rect: W.Rect): Boolean;
BEGIN
  RETURN (x >= rect.left) & (x < rect.right) & (y >= rect.top) & (y < rect.bottom);
END XYInRect;

PROCEDURE PtInRect*(rect: W.Rect; p: W.Point): Boolean;
BEGIN
  RETURN (p.x >= rect.left) & (p.x < rect.right) & (p.y >= rect.top) & (p.y < rect.bottom);
END PtInRect;

PROCEDURE LoWord*(VAR n: Longint): Longint;
BEGIN
  RETURN ASH(SYSTEM.LSH(n, 16), -16)
END LoWord;

PROCEDURE HiWord*(VAR n: Longint): Longint;
BEGIN
  RETURN ASH(n, -16);
END HiWord;

PROCEDURE Min*(a,b: Longint): Longint;
BEGIN
  IF a < b THEN
    RETURN a
  ELSE
    RETURN b;
  END;
END Min;

PROCEDURE Max*(a,b: Longint): Longint;
BEGIN
  IF a > b THEN
    RETURN a
  ELSE
    RETURN b;
  END;
END Max;

PROCEDURE Sign*(N: Longint): Longint;
VAR
  Result: Longint;
BEGIN
  IF (N > 0) THEN
    RETURN 1
  ELSIF (N < 0) THEN
    RETURN -1
  ELSE
    RETURN 0;
  END;
END Sign;

PROCEDURE Frac*(x : Real): Real;
BEGIN
  RETURN x-ENTIER(x);
END Frac;

PROCEDURE Trunc*(x : Real): Longint;
BEGIN
  RETURN ENTIER(x);
END Trunc;

PROCEDURE Floor*(x : Real) : Longint;
VAR
  Fl: Longint;
BEGIN
  Fl:=Trunc(x);
  IF Frac(x)<0 THEN Fl := Fl-1; END;
  RETURN Fl;
END Floor;

PROCEDURE Round*(x : Real): Longint;
BEGIN
  IF Frac(x)>0.5 THEN
    RETURN ENTIER(x)+1;
  ELSE
    RETURN ENTIER(x);
  END;
END Round;

PROCEDURE Ceil*(x : Real): Longint;
VAR
  Cl: Longint;
BEGIN
  Cl:=Trunc(x);
  IF Frac(x)>0 THEN Cl:=Cl+1; END;
  RETURN Cl;
END Ceil;

PROCEDURE Random*(n: Longint): Longint;
CONST
  a   = 16807;
  q   = 127773;
  r   = 2836;
  max = 2147483647;
VAR
  t: Longint;
BEGIN
  IF n <= 0 THEN n := MAX(Longint) END;
  t := a * (ZRand MOD q) - r * (ZRand DIV q);
  IF t > 0 THEN ZRand := t ELSE ZRand := t + max END;
  RETURN ZRand MOD n;
END Random;

PROCEDURE RandRange*(n1,n2: Longint): Longint;
VAR
  t: Longint;
BEGIN
  IF n1 < 0 THEN n1 := 0 END;
  IF n2 < n1 THEN n2 := n1 + 1 END;
  RETURN Random(n2-n1)+n1;
END RandRange;

PROCEDURE Length* (s: String): Longint;
VAR
  i: Longint;
BEGIN
  i := 0;
  WHILE (i < LEN(s)) & (s[i] # 0X) DO i := i + 1 END;
  RETURN i;
END Length;

PROCEDURE InsertChar*(source: Char; pos: Longint; VAR dest: String);
VAR
  dstlen, i: Longint;
BEGIN
  dstlen := Length(dest);
  IF (pos >= 0) & (pos <= dstlen) THEN
    IF dstlen < LEN(dest) - 1 THEN
      IF pos = dstlen THEN
        dest[pos+1] := 0X;
      ELSIF pos < dstlen THEN
        i := dstlen;
        WHILE i >= pos DO
          dest[i+1] := dest[i];
          DEC (i)
        END;
      END;
      dest[pos] := source;
    END;
  END;
END InsertChar;

PROCEDURE AppendChar*(source: Char; VAR dest: String);
BEGIN
  InsertChar(source,Length(dest),dest);
END AppendChar;

PROCEDURE Insert* (source: String; pos: Longint; VAR dest: String);
VAR
  srclen, dstlen, i, k : Longint;
BEGIN
  dstlen := Length(dest);
  IF (pos >= 0) & (pos <= dstlen) THEN
    IF pos = dstlen THEN
      i := 0;
      WHILE (pos + i < LEN(dest) - 1) & (source[i] # 0X) DO
        dest[pos + i] := source[i];
        i := i + 1;
      END;
      dest[pos + i] := 0X;
    ELSE
      srclen := Length(source);
      IF srclen > 0 THEN
        k := dstlen + srclen - 1;
        IF k > LEN(dest) - 2 THEN k := LEN(dest) - 2 END;
        dest[k + 1] := 0X; i := k;
        WHILE i - srclen >= pos DO
          dest[i] := dest[i - srclen];
          i := i - 1;
        END;
        i := 0;
        WHILE (i < srclen) & (pos + i <= k) DO
          dest[pos + i] := source[i];
          i := i + 1;
        END;
      END;
    END;
  END;
END Insert;

PROCEDURE Delete* (VAR s: String; pos, n : Longint);
VAR
  i, len : Longint;
BEGIN
  len := Length(s);
  IF (pos >= 0) & (pos < len) THEN
    WHILE n < len - pos DO
      s[pos] := s[pos + n];
      pos := pos + 1;
    END;
    s[pos] := 0X;
  ELSIF pos = len THEN
  END
END Delete;

PROCEDURE Replace* (source: String; pos : Longint; VAR dest: String);
BEGIN
  Delete (dest, pos, Length (source));
  Insert (source, pos, dest);
END Replace;

PROCEDURE Extract* (source: String; pos, n: Longint; VAR dest: String);
VAR
  i, len : Longint;
BEGIN
  len := Length(source);
  IF (pos >= 0) & (pos < len) THEN
    i := 0;
    IF n > len - pos THEN n := len - pos END;
    WHILE (i < n) & (i < LEN(dest) - 1) DO
      dest[i] := source[pos + i];
      i := i + 1;
    END;
    dest[i] := 0X;
  END
END Extract;

PROCEDURE Append* (extra: String; VAR dest: String);
BEGIN
  Insert (extra, Length(dest), dest);
END Append;

PROCEDURE Prepend* (VAR extra: String; dest: String);
BEGIN
  Insert (dest, Length(extra), extra);
END Prepend;

PROCEDURE Cap*(VAR s: String);
VAR
  i : Longint;
BEGIN
  i := 0;
  WHILE (i < LEN(s)) & (s[i] # 0X) DO
    IF (s[i] >= 'a') & (s[i] <= 'z') THEN s[i] := CHR(ORD(s[i]) - 20H) END;
    i := i + 1;
  END
END Cap;

PROCEDURE Copy*(src: String; VAR dest: String);
VAR
  i,len: Longint;
BEGIN
  len := LEN(src);
  IF len > LEN(dest) THEN len := LEN(dest) END;
  i := 0;
  WHILE (i < len) & (src[i] # 0X) DO
    dest[i] := src[i];
    INC(i);
  END;
  dest[i] := 0X;
END Copy;

PROCEDURE IntToStr*(x: Longint): String;
VAR
  neg  : Boolean;
  s,str: ARRAY 21 OF Char;
  i,j  : Longint;
BEGIN
  IF x # MIN(Longint) THEN

    i := 0;
    j := 0;

    IF x < 0 THEN
      neg := TRUE;
      x := -x
    ELSE
      neg := FALSE
    END;

    REPEAT
      s[i] := CHR(x MOD 10 + ORD('0'));
      INC (i);
      x := x DIV 10;
    UNTIL x = 0;

    IF ~neg THEN
      WHILE (j < i) & (j < LEN(str)) DO
        str[j] := s[i-1-j];
        INC (j);
      END;
      str[j] := 0X;
    ELSE
      str[0] := '-';
      WHILE (j < i) & (j < LEN(str)-1) DO
        str[j+1] := s[i-1-j];
        INC (j);
      END;
      str[j+1] := 0X;
    END
  ELSE
    str[0] := 0X;
    Append('-2147483648',str);
  END;
  RETURN str;
END IntToStr;

PROCEDURE HexToStr*(x: Longint): String;
VAR
  s,str: ARRAY 9 OF Char;
  i, j, n: Longint;
BEGIN
  i := 0;
  REPEAT
    n := x MOD 16;
    IF n < 10 THEN
      s[i] := CHR(n + ORD('0'));
    ELSE
      s[i] := CHR(ORD('A') - 10 + n);
    END;
    INC (i);
    x := x DIV 16;
  UNTIL (x = 0) OR (i = 8);
  j := 0;
  WHILE (j < i) & (j < LEN(str)) DO
    str[j] := s[i-1-j];
    INC (j)
  END;
  str[j] := 0X;
  RETURN str;
END HexToStr;

PROCEDURE Lowercase*(name: String; VAR result: String);
VAR
  ch: Char;
  i : Longint;
BEGIN
  i := 0;
  REPEAT
    ch := name[i];
    IF (ch >= 'A') & (ch <= 'Z') THEN
      ch := CHR(ORD(ch)-ORD('A')+ORD('a'));
    END;
    result[i] := ch;
    INC(i);
  UNTIL ch = 0X;
END Lowercase;

PROCEDURE Uppercase*(name: String; VAR result: String);
VAR
  ch: Char;
  i : Longint;
BEGIN
  i := 0;
  REPEAT
    ch := name[i];
    IF (ch >= 'a') & (ch <= 'z') THEN
      ch := CHR(ORD(ch)-ORD('a')+ORD('A'));
    END;
    result[i] := ch;
    INC(i);
  UNTIL ch = 0X;
END Uppercase;

PROCEDURE Align*(VAR offset: Longint; alignment: Longint);
BEGIN
  IF alignment >0  THEN
    INC(offset,(-offset) MOD alignment);
  ELSIF alignment < 0 THEN
    DEC(offset,offset MOD (-alignment));
  END;
END Align;

BEGIN
  ZRand := W.GetTickCount();
END Utilites.
